from .base import \
    NeuralNetworkModule, \
    dynamic_module_wrapper, \
    static_module_wrapper

from .resnet import ResNet

from . import base
from . import resnet

__all__ = ["NeuralNetworkModule",
           "dynamic_module_wrapper",
           "static_module_wrapper",
           "ResNet",
           "base",
           "resnet"]



from enum import Enum
from time import sleep
from datetime import datetime
from queue import Queue, Empty
from abc import ABC, abstractmethod
from threading import Thread, Condition, Lock
from typing import Any, Dict, Union, List, Callable

from torch.distributed import rpc
from machin.utils.helper_classes import Timer, Counter


class ElectionGroupBase(ABC):
    """
    Base class for all rpc group election algorithms and their implementations.

    Attributes:
        leader_change_cond: An election group should call ``notify_all()`` on
            this conditional if a leader change event has happened. Services
            relying on the leader may create a task thread waiting on this
            conditional.
    """
    def __init__(self):
        self.leader_change_cond = Condition(Lock())

    @abstractmethod
    def start(self):
        """
        Start the election algorithm.
        """
        pass

    @abstractmethod
    def get_leader(self) -> Union[int, None]:
        """
        Returns:
            Absolute global process rank of the current leader.

            If no leader has been elected, return ``None``.
        """
        pass

    @abstractmethod
    def is_leader(self) -> bool:
        """
        Returns:
            Whether current process is the leader.
        """
        pass

    def get_leader_change_cond(self):
        """
        Warning:
            This conditional is allowed to notify the same leader
            change event for indefinite times.

        Returns:
            The leader change conditional.
        """
        return self.leader_change_cond

    @abstractmethod
    def get_members(self) -> List[int]:
        """
        Returns:
            A list of the absolute global rank of all current members.
        """
        pass


class ElectionGroupSimple(ElectionGroupBase):
    """
    The simplest static election group, leader is specified on group creation.
    And will remain the same in the whole life time of the group. Requires no
    communication and adds zero overhead.
    """
    def __init__(self,
                 member_ranks: List[int],
                 rank: int,
                 leader_rank: int = 0,
                 **__):
        """
        Args:
            member_ranks: Absolute member process ranks.
            rank: Rank of the current process.
            leader_rank: Absolute leader process rank
        """
        self.member_ranks = sorted(member_ranks)
        self.leader_rank = leader_rank
        self.rank = rank
        super(ElectionGroupSimple, self).__init__()

    def start(self):
        # DOC INHERITED
        pass

    def get_leader(self):
        # DOC INHERITED
        return self.leader_rank

    def is_leader(self):
        # DOC INHERITED
        return self.rank == self.leader_rank

    def get_members(self):
        # DOC INHERITED
        return self.member_ranks


class MType(Enum):
    """
    Message types used in
    :class:`~machin.parallel.distributed.election.ElectionGroupStableBase`
    """
    START = 0
    OK = 1
    ALERT = 2
    PING = 3
    PONG = 4


class ElectionGroupStableBase(ElectionGroupBase):
    """
    This class implements the stable election algorithm described in:
    `<<Stable leader election>> <http://citeseerx.ist.psu.edu/viewdoc/\
download?doi=10.1.1.89.4817&rep=rep1&type=pdf>`_

    The essay stated that the algorithm can establish leadership within
    :math:`9\\delta` time. The algorithm requires :math:`O(n^2)` messages
    to elect and :math:`O(n)` messages to maintain. But the leadership
    is stable, so no scenarios like leadership switching back and forth
    between two nodes will ocurr, this is favorable.

    Code in this class implements the algorithm described in Figure 4 with
    optimizations in Figure 5 of the essay.
    """
    # The ratio of timeout value to sleep time between probes.
    SLEEP_DIVISION = 1000

    # Stores references to all stable election groups, rpc calls
    # relies on this dictionary to find the target election group/

    # This is better than using a global variable, since its name scope
    # is limited within this class.
    elect_groups = {}

    def __init__(self,
                 name: str,
                 member_ranks: List[int],
                 rank: int,
                 timeout: float = 1,
                 **__):
        """
        Args:
            name: Name of this election group.
            member_ranks: Absolute member ranks.
            rank: Rank of the current process.
            timeout: Timeout :math:`\\delta` used in the essay.
        """
        super(ElectionGroupStableBase, self).__init__()
        if name in self.elect_groups:
            raise RuntimeError("Election group already existed!")
        self.member_ranks = sorted(member_ranks)

        # relative rank in list
        self.rank = self.member_ranks.index(rank)
        # Name of group, used to lookup
        self.name = name

        self.timer = Timer()
        self.alive_timer = Timer()
        self.timeout_recv_timer = Timer()
        self.timeout_recv_ok_or_start = False
        self.timeout_recv_pong = []
        self.timeout = timeout

        self.cur_round = 0
        self.leader = None

        self.tka_thread = Thread(target=self._task_keep_alive)
        self.tto_thread = Thread(target=self._task_timeout)
        self.leader_change_cond = Condition(Lock())
        self.ok_counter = Counter()
        self.last_alert = None
        self.recv_alerts = []
        self.elect_groups[name] = self

        self.stop = False
        self.tka_thread.start()
        self.tto_thread.start()
        self._start_round(0)
        self._register_handle(self._handle)

    def start(self):
        # DOC INHERITED
        self.tka_thread.start()
        self.tto_thread.start()

    def get_leader(self):
        # DOC INHERITED

        # In order to deal with leader ship changes during
        # member rank lookup, we should fetch the leader first

        # If leadership has changed after fetch, we may use
        # leader_change_cond to detect the change.
        leader = self.leader
        if leader is None:
            return None
        return self.member_ranks[leader]

    def is_leader(self):
        # DOC INHERITED
        return self.rank == self.leader

    def get_members(self):
        # DOC INHERITED
        return self.member_ranks

    def _start_round(self, cur_round):
        # The start round function defined in essay.
        self._send_all((MType.ALERT, cur_round))
        if self.rank != cur_round % len(self.member_ranks):
            self._send_all((MType.START, cur_round))
        self.cur_round = cur_round
        self.leader = None
        self.ok_counter.reset()
        self.timer.begin()
        self.leader_change_cond.notify_all()

    def _handle(self, timestamp, src, message):
        # The handle task defined in essay.
        if message == (MType.OK, self.cur_round):
            self.ok_counter.count()
            if (self.leader is None and
                    self.ok_counter >= 2 and
                    (self.last_alert is None or
                     self._timestamp() - timestamp > 6 * self.timeout)):
                self.leader = self.cur_round % len(self.member_ranks)
                self.timer.begin()
                self.leader_change_cond.notify_all()

        elif (message[0] in (MType.OK, MType.START) and
              message[1] > self.cur_round):
            self._start_round(self.cur_round + 1)
            self.timeout_recv_ok_or_start = True

        elif (message[0] in (MType.OK, MType.START) and
              message[1] < self.cur_round):
            self._send(src, (MType.START, self.cur_round))

        elif (message[0] == MType.ALERT and
              message[1] > self.cur_round):
            self.leader = None
            self.last_alert = self._timestamp()
            self.recv_alerts.append((timestamp, message))
            self.leader_change_cond.notify_all()

        elif message[0] == MType.PING:
            self._send(src, (MType.PONG, message[1]))

        elif message == (MType.PONG, self.cur_round):
            self.timeout_recv_pong.append(src)

    def _task_timeout(self):
        # The Figure 5 optimization in essay.
        while not self.stop:
            if self.timer.end() > self.timeout * 2:
                self._send_all((MType.ALERT, self.cur_round + 1))

                self.timeout_recv_pong.clear()
                self._send_all((MType.PING, self.cur_round))

                self.timeout_recv_ok_or_start = False
                self.timeout_recv_timer.begin()

                while (not self.timeout_recv_ok_or_start and
                       self.timeout_recv_timer.end() <= 2 * self.timeout):
                    sleep(self.timeout / self.SLEEP_DIVISION)
                if self.timeout_recv_ok_or_start:
                    continue

                # Jump to the nearest available round
                # (and the respective leader candidate),
                # skipping all not-responding nodes to save time.
                cur_leader_rank = self.cur_round % len(self.member_ranks)
                avail_ranks = self.timeout_recv_pong.copy()
                for avail_rank in avail_ranks:
                    if avail_rank > cur_leader_rank:
                        new_round = self.cur_round + \
                                    (avail_rank - cur_leader_rank)
                        self._start_round(new_round)
                        break
                else:
                    new_round = min(avail_ranks) + len(self.member_ranks) - \
                                cur_leader_rank + self.cur_round
                    self._start_round(new_round)
            else:
                sleep(self.timeout / self.SLEEP_DIVISION)

    def _task_keep_alive(self):
        # The keep alive task defined in essay.
        while not self.stop:
            if self.rank == self.cur_round % len(self.member_ranks):
                if self.alive_timer.end() >= self.timeout:
                    self._send_all((MType.OK, self.cur_round))
                    self.alive_timer.begin()
            sleep(self.timeout / self.SLEEP_DIVISION)

    @abstractmethod
    def _register_handle(self, handler: Callable):
        # Register the _handle() function, since some communication
        # libraries use callback handlers to deal with incoming messages.
        pass

    @abstractmethod
    def _send(self, to: int, message: Any):
        # Actual implementations must implement this transmission function.
        # ``to`` is the absolute process rank.
        pass

    @abstractmethod
    def _send_all(self, message: Any):
        # Actual implementations must implement this transmission function.
        pass

    @staticmethod
    def _timestamp():
        return datetime.utcnow().timestamp()

    def __del__(self):
        # Close this group.
        self.stop = True
        self.tka_thread.join()
        self.tto_thread.join()


class ElectionGroupStableRpc(ElectionGroupStableBase):
    # Implements the transmission layer using RPC.
    def _register_handle(self, handler):
        self.h_thread = Thread(target=self._task_handle)
        self.recv_queue = Queue()

    def _send(self, to, message):
        ts = self._timestamp()
        to = self.member_ranks[to]
        rpc.rpc_async(str(to), self._reply,
                      args=(ts, self.rank, self.name, message))

    def _send_all(self, message):
        ts = self._timestamp()
        for to in self.member_ranks:
            rpc.rpc_async(str(to), self._reply,
                          args=(ts, self.rank, self.name, message))

    def _task_handle(self):
        while not self.stop:
            try:
                packet = self.recv_queue.get(block=True, timeout=1e-3)
            except Empty:
                pass
            else:
                self._handle(*packet)

    def _recv(self, timestamp, src, message):
        # should only be called by _reply
        self.recv_queue.put_nowait((timestamp, src, message))

    @staticmethod
    def _reply(timestamp: float, src: int, group: str, message: Any):
        # This function is called by rpc api to finish delivering
        # the message.
        elect_groups = \
            ElectionGroupStableRpc\
            .elect_groups  # type: Dict[str, ElectionGroupStableRpc]
        timeout = elect_groups[group].timeout

        # pylint: disable=protected-access
        # Actually it is accessing its own _recv method, not really protected.
        if datetime.utcnow().timestamp() - timestamp <= timeout:
            elect_groups[group]._recv(timestamp, src, message)



from .a2c import *


class PPO(A2C):
    """
    PPO framework.
    """
    def __init__(self,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 *_,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
                 learning_rate: float = 0.001,
                 entropy_weight: float = None,
                 value_weight: float = 0.5,
                 surrogate_loss_clip: float = 0.2,
                 gradient_max: float = np.inf,
                 gae_lambda: float = 1.0,
                 discount: float = 0.99,
                 update_times: int = 50,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 visualize: bool = False,
                 **__):
        """
        See Also:
            :class:`.A2C`

        Args:
            actor: Actor network module.
            critic: Critic network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            criterion: Critierion used to evaluate the value loss.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            entropy_weight: Weight of entropy in your loss function, a positive
                entropy weight will minimize entropy, while a negative one will
                maximize entropy.
            value_weight: Weight of critic value loss.
            surrogate_loss_clip: :math:`\\epsilon` used in surrogate loss
                clipping:
                :math:`L_t(\\theta)=min(r_t(\\theta)\\hat{A_t}, \
                                        clip(r_t(\\theta), \
                                             1-\\epsilon, 1+\\epsilon) \
                                        \\hat{A_t})`
            gradient_max: Maximum gradient.
            gae_lambda: :math:`\\lambda` used in generalized advantage
                estimation.
            discount: :math:`\\gamma` used in the bellman function.
            update_times: Number of update iterations per sample period. Buffer
                will be cleared after ``update()``
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            visualize: Whether visualize the network flow in the first pass.
        """
        super(PPO, self).__init__(actor, critic, optimizer, criterion,
                                  lr_scheduler=lr_scheduler,
                                  lr_scheduler_args=lr_scheduler_args,
                                  lr_scheduler_kwargs=lr_scheduler_kwargs,
                                  learning_rate=learning_rate,
                                  entropy_weight=entropy_weight,
                                  value_weight=value_weight,
                                  gradient_max=gradient_max,
                                  gae_lambda=gae_lambda,
                                  discount=discount,
                                  update_times=update_times,
                                  replay_size=replay_size,
                                  replay_device=replay_device,
                                  replay_buffer=replay_buffer,
                                  visualize=visualize)
        self.surr_clip = surrogate_loss_clip

    def update(self,
               update_value=True,
               update_policy=True,
               concatenate_samples=True,
               **__):
        # DOC INHERITED
        sum_act_policy_loss = 0
        sum_value_loss = 0

        # sample a batch
        batch_size, (state, action, reward, next_state, terminal,
                     action_log_prob, target_value, advantage, *others) = \
            self.replay_buffer.sample_batch(-1,
                                            sample_method="all",
                                            concatenate=concatenate_samples,
                                            sample_attrs=[
                                                "state", "action", "reward",
                                                "next_state", "terminal",
                                                "action_log_prob", "value",
                                                "gae", "*"],
                                            additional_concat_attrs=[
                                                "action_log_prob", "value",
                                                "gae"
                                            ])

        # normalize target value
        target_value = ((target_value - target_value.mean()) /
                        (target_value.std() + 1e-5))

        # Infer original action log probability
        __, action_log_prob, *_ = self.eval_act(state, action)
        action_log_prob = action_log_prob.view(batch_size, 1)

        for i in range(self.update_times):
            value = self.criticize(state)

            if self.entropy_weight is not None:
                __, new_action_log_prob, new_action_entropy, *_ = \
                    self.eval_act(state, action)

            else:
                __, new_action_log_prob, *_ = self.eval_act(state, action)
                new_action_entropy = None

            new_action_log_prob = new_action_log_prob.view(batch_size, 1)

            # calculate surrogate loss
            sim_ratio = t.exp(new_action_log_prob - action_log_prob).detach()
            advantage = advantage.to(sim_ratio.device)
            surr_loss_1 = sim_ratio * advantage
            surr_loss_2 = t.clamp(sim_ratio,
                                  1 - self.surr_clip,
                                  1 + self.surr_clip) * advantage

            # calculate policy loss using surrogate loss
            act_policy_loss = -t.min(surr_loss_1, surr_loss_2)

            if new_action_entropy is not None:
                act_policy_loss += (self.entropy_weight *
                                    new_action_entropy.mean())

            act_policy_loss = act_policy_loss.mean()

            value_loss = (self.criterion(target_value.to(value.device), value) *
                          self.value_weight)

            if self.visualize:
                self.visualize_model(act_policy_loss, "actor")

            # Update actor network
            if update_policy:
                self.actor.zero_grad()
                act_policy_loss.backward()
                nn.utils.clip_grad_norm_(
                    self.actor.parameters(), self.grad_max
                )
                self.actor_optim.step()
                sum_act_policy_loss += act_policy_loss.item()

            if self.visualize:
                self.visualize_model(value_loss, "critic")

            # Update critic network
            if update_value:
                self.critic.zero_grad()
                value_loss.backward()
                nn.utils.clip_grad_norm_(
                    self.critic.parameters(), self.grad_max
                )
                self.critic_optim.step()
                sum_value_loss += value_loss.item()

        self.replay_buffer.clear()
        return (-sum_act_policy_loss / self.update_times,
                sum_value_loss / self.update_times)



import os
import time
import torch
import torch.nn as nn
import numpy as np
from torch.multiprocessing import Process
import torch.distributed.rpc as rpc

rank = None
messages = 225
world_size = 8


def test():
    return "pong, {}, {}, {}, {}, {}"
    #return "pong, {}, {}, {}, {}, {}".format(rank, rank, rank, rank, rank)


@torch.jit.script
def test1():
    return "pong, {}, {}, {}, {}, {}"


def run(i):
    rpc.init_rpc("Rank"+str(i), rank=i, world_size=world_size,
                 rpc_backend_options=rpc.ProcessGroupRpcBackendOptions(
                     init_method="env://",
                     rpc_timeout=rpc.timedelta(seconds=60),
                     num_send_recv_threads=4))
    t = time.time()
    reqs = []
    for j in range(messages):
        for r in range(world_size):
            reqs.append(rpc.rpc_async("Rank{}".format(r), test, args=()))
    for req in reqs:
        req.wait()
    print(time.time() - t)
    rpc.shutdown()


if __name__ == "__main__":
    os.environ['MASTER_ADDR'] = "localhost"
    os.environ['MASTER_PORT'] = "29500"
    ps = []
    for i in range(world_size):
        p = Process(target=run, args=(i,))
        p.start()
        ps.append(p)

    for p in ps:
        p.join()



from abc import ABC, abstractmethod
from typing import Union, List, Any


class ParallelWrapperBase(ABC):
    def __init__(self, *_, **__):
        pass

    @abstractmethod
    def reset(self, idx: Union[int, List[int], None] = None) -> Any:
        """
        Reset all environments if id is ``None``, otherwise reset the specific
        environment(s) with given index(es).

        Args:
            idx: Environment index(es) to be reset.

        Returns:
            Initial observation of all environments. Format is unspecified.
        """
        pass

    @abstractmethod
    def step(self, action, idx: Union[int, List[int], None] = None) -> Any:
        """
        Let specified environment(s) run one time step. specified environments
        must be active and have not reached terminal states before.

        Args:
            action: actions to take.
            idx: Environment index(es) to be run.

        Returns:
            New states of environments.
        """
        pass

    @abstractmethod
    def seed(self, seed: Union[int, List[int], None] = None) -> List[int]:
        """
        Set seed(s) for all environment(s).

        Args:
            seed: A single integer seed for all environments,
                or a list of integers for each environment,
                or None for default seed.

        Returns:
            New seed of each environment.
        """
        pass

    @abstractmethod
    def render(self, *args, **kwargs) -> Any:
        """
        Render all environments.
        """
        pass

    @abstractmethod
    def close(self) -> Any:
        """
        Close all environments.
        """
        pass

    @abstractmethod
    def active(self) -> List[int]:
        """
        Returns: Indexes of active environments.
        """
        pass

    @abstractmethod
    def size(self) -> int:
        """
        Returns:
             Number of environments.
        """
        pass



import torch as t
import numpy as np
import torch.nn as nn
from torch.nn.utils.weight_norm import weight_norm

from .base import NeuralNetworkModule


def conv3x3(in_planes, out_planes, stride=1):
    """
    Create a 3x3 2d convolution block
    """
    return nn.Conv2d(in_planes, out_planes,
                     kernel_size=3, stride=stride, padding=1, bias=False)


def conv5x5(in_planes, out_planes, stride=2):
    """
    Create a 5x5 2d convolution block
    """
    return nn.Conv2d(in_planes, out_planes,
                     kernel_size=5, stride=stride, padding=2, bias=False)


def cfg(depth, use_batch_norm=True):
    depth_lst = [18, 34, 50, 101, 152]
    if depth not in depth_lst:
        raise ValueError("Error : Resnet depth should be either "
                         "18, 34, 50, 101, 152")
    if use_batch_norm:
        cfg_dict = {
            '18': (BasicBlock, [2, 2, 2, 2]),
            '34': (BasicBlock, [3, 4, 6, 3]),
            '50': (Bottleneck, [3, 4, 6, 3]),
            '101': (Bottleneck, [3, 4, 23, 3]),
            '152': (Bottleneck, [3, 8, 36, 3]),
        }
    else:
        cfg_dict = {
            '18': (BasicBlockWN, [2, 2, 2, 2]),
            '34': (BasicBlockWN, [3, 4, 6, 3]),
            '50': (BottleneckWN, [3, 4, 6, 3]),
            '101': (BottleneckWN, [3, 4, 23, 3]),
            '152': (BottleneckWN, [3, 8, 36, 3]),
        }
    return cfg_dict[str(depth)]


class BasicBlock(NeuralNetworkModule):
    # Expansion parameter, output will have "expansion * in_planes" depth.
    expansion = 1

    def __init__(self, in_planes, out_planes, stride=1):
        """
        Create a basic block of resnet.

        Args:
            in_planes:  Number of input planes.
            out_planes: Number of output planes.
            stride:     Stride of convolution.
        """
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(in_planes, out_planes, stride)
        self.bn1 = nn.BatchNorm2d(out_planes)
        self.conv2 = conv3x3(out_planes, out_planes)
        self.bn2 = nn.BatchNorm2d(out_planes)
        # Create a shortcut from input to output.
        # An empty sequential structure means no transformation
        # is made on input X.
        self.shortcut = nn.Sequential()

        self.set_input_module(self.conv1)

        # A convolution is needed if we cannot directly add input X to output
        # BatchNorm2d produces NaN gradient?
        if stride != 1 or in_planes != self.expansion * out_planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * out_planes,
                          kernel_size=1, stride=stride, bias=False),
                # nn.BatchNorm2d(self.expansion * out_planes)
            )

    def forward(self, x):
        out = t.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = t.relu(out)
        return out


class Bottleneck(NeuralNetworkModule):
    # Expansion parameter, output will have "expansion * in_planes" depth.
    expansion = 4

    def __init__(self, in_planes, out_planes, stride=1):
        """
        Create a bottleneck block of resnet.

        Args:
            in_planes:  Number of input planes.
            out_planes: Number of output planes.
            stride:     Stride of convolution.
        """
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, out_planes,
                               kernel_size=1, bias=False)
        self.conv2 = nn.Conv2d(out_planes, out_planes,
                               kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.conv3 = nn.Conv2d(out_planes, self.expansion * out_planes,
                               kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(out_planes)
        self.bn2 = nn.BatchNorm2d(out_planes)
        self.bn3 = nn.BatchNorm2d(self.expansion * out_planes)

        self.shortcut = nn.Sequential()

        self.set_input_module(self.conv1)

        if stride != 1 or in_planes != self.expansion * out_planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * out_planes,
                          kernel_size=1, stride=stride, bias=False),
            )

    def forward(self, x):
        out = t.relu(self.bn1(self.conv1(x)))
        out = t.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        out = t.relu(out)
        return out


class BasicBlockWN(NeuralNetworkModule):
    """
    Basic block with weight normalization
    """
    # Expansion parameter, output will have "expansion * in_planes" depth.
    expansion = 1

    def __init__(self, in_planes, out_planes, stride=1):
        """
        Create a basic block of resnet.

        Args:
            in_planes:  Number of input planes.
            out_planes: Number of output planes.
            stride:     Stride of convolution.
        """
        super(BasicBlockWN, self).__init__()
        self.conv1 = weight_norm(conv3x3(in_planes, out_planes, stride))
        self.conv2 = weight_norm(conv3x3(out_planes, out_planes))
        # Create a shortcut from input to output.
        # An empty sequential structure means no transformation
        # is made on input X.
        self.shortcut = nn.Sequential()

        self.set_input_module(self.conv1)

        # A convolution is needed if we cannot directly add input X to output.
        if stride != 1 or in_planes != self.expansion * out_planes:
            self.shortcut = nn.Sequential(
                weight_norm(
                    nn.Conv2d(in_planes, self.expansion * out_planes,
                              kernel_size=1, stride=stride, bias=False)),
            )

    def forward(self, x):
        out = t.relu(self.conv1(x))
        out = self.conv2(out)
        out += self.shortcut(x)
        out = t.relu(out)
        return out


class BottleneckWN(NeuralNetworkModule):
    """
    Bottleneck block with weight normalization
    """
    # expansion parameter, output will have "expansion * in_planes" depth
    expansion = 4

    def __init__(self, in_planes, out_planes, stride=1):
        """
        Create a bottleneck block of resnet.

        Args:
            in_planes:  Number of input planes.
            out_planes: Number of output planes.
            stride:     Stride of convolution.
        """
        super(BottleneckWN, self).__init__()
        self.conv1 = weight_norm(nn.Conv2d(in_planes, out_planes,
                                           kernel_size=1, bias=False))
        self.conv2 = weight_norm(nn.Conv2d(out_planes, out_planes,
                                           kernel_size=3, stride=stride,
                                           padding=1, bias=False))
        self.conv3 = weight_norm(nn.Conv2d(out_planes,
                                           self.expansion * out_planes,
                                           kernel_size=1, bias=False))

        self.shortcut = nn.Sequential()

        self.set_input_module(self.conv1)

        if stride != 1 or in_planes != self.expansion * out_planes:
            self.shortcut = nn.Sequential(
                weight_norm(
                    nn.Conv2d(in_planes, self.expansion * out_planes,
                              kernel_size=1, stride=stride, bias=False)),
            )

    def forward(self, x):
        out = t.relu(self.conv1(x))
        out = t.relu(self.conv2(out))
        out = self.conv3(out)
        out += self.shortcut(x)
        out = t.relu(out)

        return out


class ResNet(NeuralNetworkModule):
    def __init__(self,
                 in_planes: int,
                 depth: int,
                 out_planes: int,
                 out_pool_size=(1, 1),
                 use_batch_norm=True):
        """
        Create a resnet of specified depth.

        Args:
            in_planes:  Number of input planes.
            depth: Depth of resnet. Could be one of ``18, 34, 50, 101, 152``.
            out_planes: Number of output planes.
            out_pool_size: Size of pooling output
        """
        super(ResNet, self).__init__()
        self.in_planes = 64
        self.out_pool_size = out_pool_size

        block, num_blocks = cfg(depth, use_batch_norm)

        self.conv1 = conv3x3(in_planes, 64, 2)
        if use_batch_norm:
            self.bn1 = nn.BatchNorm2d(64)
            self.layer1 = self._make_layer(block, 64, num_blocks[0], 2)
            self.layer2 = self._make_layer(block, 128, num_blocks[1], 2)
            self.layer3 = self._make_layer(block, 256, num_blocks[2], 2)
            self.layer4 = self._make_layer(block, 512, num_blocks[3], 2)
            self.base = nn.Sequential(self.conv1, self.bn1, nn.ReLU(),
                                      self.layer1, self.layer2,
                                      self.layer3, self.layer4)
        else:
            self.layer1 = self._make_layer(block, 64, num_blocks[0], 2)
            self.layer2 = self._make_layer(block, 128, num_blocks[1], 2)
            self.layer3 = self._make_layer(block, 256, num_blocks[2], 2)
            self.layer4 = self._make_layer(block, 512, num_blocks[3], 2)
            self.base = nn.Sequential(self.conv1, nn.ReLU(),
                                      self.layer1, self.layer2,
                                      self.layer3, self.layer4)
        self.fc = nn.Linear(512 * out_pool_size[0] * out_pool_size[1],
                            out_planes)

        self.set_input_module(self.conv1)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []

        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.base(x)
        kernel_size = (np.int(np.floor(x.size(2) / self.out_pool_size[0])),
                       np.int(np.floor(x.size(3) / self.out_pool_size[1])))
        x = nn.functional.avg_pool2d(x, kernel_size)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        return x



from abc import ABC, abstractmethod
from collections import OrderedDict
from threading import Lock
from random import choice
from machin.parallel.distributed import RpcGroup


class OrderedServerBase(ABC):
    @abstractmethod
    def push(self, key, value, version, prev_version):
        """
        Push a new ``version`` of ``value`` in ``key`` to the ordered server.

        Note:
            If ``version = prev_version`` then there is no order guarantee. But
            you may exploit this feature.

        Args:
            key: Key.
            value: value.
            version: New version.
            prev_version: Previous version.

        Returns:
            ``True`` if success, and ``False`` if not.
        """
        pass

    @abstractmethod
    def pull(self, key, version=None):
        """
        Pull a value with the specified ``version`` in ``key``.

        Args:
            key: Key.
            version: Target version, if ``None``, then the newest version
                of value of key will be pulled.

        Returns:
            ``None`` if version is not found, auto-deleted, or key is not found,
            otherwise returns value with the specified ``version``
            in ``key``, and then ``version``
        """
        pass


class SimpleOrderedServer(OrderedServerBase):
    """
    A simple key-value server, with strict ordered update and automatic
    replication.

    Note:
        This simple implementation is based on a master-slave architecture,
        and only provides availability, but provides no guarantee on
        consistency, if master fails, the whole server group
        will fail.
    """
    def __init__(self,
                 group: RpcGroup,
                 master: int = 0,
                 version_depth: int = 1,
                 **__):
        """
        Args:
            group: The rpc group of this server.
            master: The relative rpc rank of the master server. Must
                be less than group size.
            version_depth: Storage depth of old versions of the same
                key. If ``depth = 1``, then only the newest version
                of the key will be saved.
        """
        self.group = group
        self.group.rpc_register_paired(self.__class__, self)
        self.storage = {}
        self.lock = Lock()
        self.log_depth = version_depth
        self.master = self.group.get_group_members()[master]
        self.slaves = [member
                       for member in self.group.get_group_members()
                       if member != self.master]

        assert version_depth > 0 and isinstance(version_depth, int)

    def push(self, key, value, version, prev_version):
        # DOC INHERITED
        return self.group.rpc_paired_class_sync(self.master,
                                                self._reply_push,
                                                self.__class__,
                                                args=(key, value, version,
                                                      prev_version))

    def pull(self, key, version=None):
        # DOC INHERITED
        to = choice(self.group.get_group_members())
        return self.group.rpc_paired_class_sync(to,
                                                self._reply_pull,
                                                self.__class__,
                                                args=(key, version))[1]

    def _reply_push(self, key, value, version, prev_version):
        success = False
        self.lock.acquire()
        if key in self.storage:
            ref = self.storage[key]
            # Check previous version consistency.
            if next(reversed(ref)) == prev_version:
                ref[version] = value
                success = True
            if len(ref) > self.log_depth + 1:
                ref.pop(0)
        else:
            # Create a new key.
            ref = self.storage[key] = OrderedDict()
            ref[version] = value
            success = True

        future = []
        for slave in self.slaves:
            # push updates to slaves
            future.append(self.group.rpc_paired_class_sync(
                slave, self._master_sync, self.__class__,
                args=(key, value, version)
            ))

        self.lock.release()
        return success

    def _reply_pull(self, key, version):
        result = None
        self.lock.acquire()
        if key in self.storage:
            ref = self.storage[key]
            # Try to find the target version.
            if version is not None and version in ref:
                result = (ref[version], version)
            # Find the newest version.
            elif version is None:
                result = (ref[-1], next(reversed(ref)))
        self.lock.release()
        return result

    def _master_sync(self, key, value, version):
        # faster _reply_push, for master-slave sync
        if key in self.storage:
            ref = self.storage[key]
            ref[version] = value
            if len(ref) > self.log_depth + 1:
                ref.pop(0)
        else:
            ref = self.storage[key] = OrderedDict()
            ref[version] = value



from typing import Any
from multiprocessing import context, connection
import sys
import dill
import torch as t

from .pickle import dump_tensor


class ConnectionWrapper(object):
    """
    This simple wrapper provides timeout function for sending
    bytes on ``Connection``.
    """

    def __init__(self, conn):
        self.conn = conn

    def send_bytes(self, bytes_):
        """
        Send bytes over the connection pipe.
        """
        self.conn.send_bytes(bytes_)

    def recv_bytes(self, timeout=None):
        """
        Receive bytes from the connection pipe.

        Raises:
            TimeoutError if timeout.
        """
        if self.conn.poll(timeout=timeout):
            return self.conn.recv_bytes()
        else:
            raise TimeoutError("Timeout")

    def __getattr__(self, name):
        if 'conn' in self.__dict__:
            return getattr(self.conn, name)
        raise AttributeError("'{}' object has no attribute '{}'".format(
            type(self).__name__, 'conn'))


class SimpleQueue(object):
    """
    A simple queue for inter-process communications.
    """
    def __init__(self, *, ctx=None, copy_tensor=False):
        """
        Args:
            ctx: Multiprocessing context, you can get this using ``get_context``
            copy_tensor: Set the queue to send a fully serialized tensor
                if ``True``, and only a stub of reference if ``False``.

        See Also:
            :func:`.dump_tensor`
        """

        self._reader, self._writer = connection.Pipe(duplex=False)
        self._reader = ConnectionWrapper(self._reader)
        self._writer = ConnectionWrapper(self._writer)
        self._r_lock = ctx.Lock()
        self._poll = self._reader.poll
        self._copy_tensor = copy_tensor
        if sys.platform == 'win32':
            self._w_lock = None
        else:
            self._w_lock = ctx.Lock()

    def empty(self):
        """
        Returns:
            Whether the queue is empty or not.
        """
        return not self._poll()

    def __getstate__(self):
        context.assert_spawning(self)
        return self._reader, self._writer, self._r_lock, self._w_lock

    def __setstate__(self, state):
        (self._reader, self._writer, self._r_lock, self._w_lock) = state

    def get(self):
        """
        Get an object from the queue. This api is required by
        ``multiprocessing.pool`` to perform inter-process
        communication.

        Note:
            This api is used by sub-processes in pool to get tasks
            and work.

        Returns:
            Any object.
        """
        with self._r_lock:
            res = self._reader.recv_bytes()
        # deserialize the data after having released the lock
        return dill.loads(res)

    def put(self, obj: Any):
        """
        Put an object into the queue. This api is required by
        ``multiprocessing.pool`` to perform inter-process
        communication.

        Note:
            This api is used by sub-processes in pool to put results
            and respond.

        Args:
            obj: Any object.
        """
        # serialize the data before acquiring the lock
        if t.is_tensor(obj):
            obj = dump_tensor(obj, self._copy_tensor)
        else:
            obj = dill.dumps(obj)
        if self._w_lock is None:
            # writes to a message oriented win32 pipe are atomic
            self._writer.send_bytes(obj)
        else:
            with self._w_lock:
                self._writer.send_bytes(obj)

    def quick_get(self):
        """
        Get an object from the queue.

        Note:
            this api is used by the result manager (``Pool._result_handler``)
            thread to get results from the queue, since it is single threaded,
            there is no need to use locks, and therefore quicker.

        Returns:
            Any object.
        """
        res = self._reader.recv_bytes()
        return dill.loads(res)

    def quick_put(self, obj: Any):
        """
        Put an object into the queue.

        Note: this api is used by the pool manager (``Pool._task_handler``)
            thread to put tasks into the queue, since it is single threaded,
            there is no need to use locks, and therefore quicker.

        Args:
            obj: Any object.
        """
        if t.is_tensor(obj):
            obj = dump_tensor(obj, self._copy_tensor)
        else:
            obj = dill.dumps(obj)
        self._writer.send_bytes(obj)


__all__ = ["SimpleQueue"]



"""
Attributes:
    default_board: The default global board.
"""
import numpy as np
from tensorboardX import SummaryWriter


class TensorBoard:
    """
    Create a tensor board object.

    Attributes:
        writer: ``SummaryWriter`` of package ``tensorboardX``.
    """
    def __init__(self):
        self.writer = None

    def init(self, *writer_args):
        if self.writer is None:
            self.writer = SummaryWriter(*writer_args)
        else:
            raise RuntimeError("Writer has been initialized!")

    def is_inited(self) -> bool:
        """
        Returns: whether the board has been initialized with a writer.
        """
        return not self.writer is None


def normalize_seq_length(seq, length):
    return np.tile(
        np.array(seq),
        int(np.ceil(length / len(seq)))
    )[:length].tolist()


default_board = TensorBoard()



from machin.utils.helper_classes import Switch
from machin.frame.noise.param_space_noise import perturb_model
from machin.utils.visualize import visualize_graph
import torch as t

dims = 5

t.manual_seed(0)
model = t.nn.Linear(dims, dims)
optim = t.optim.Adam(model.parameters(), 1e-3)
p_switch, r_switch = Switch(), Switch()
rst_func = perturb_model(model, p_switch, r_switch)
r_switch.on()

# turn off/on the perturbation switch to see the difference
p_switch.on()

# do some sampling
action = model(t.ones([dims]))

# Visualize will not show any leaf noise tensors
# because they are created in t.no_grad() context.
visualize_graph(action, exit_after_vis=False)

# do some training
loss = (action - t.ones([dims])).sum()
loss.backward()
rst_func()
optim.step()
print(model.weight)



from torchviz import make_dot
"""
The visualization module, currently it only contains the pytorch
flow graph visualization, more visualizations for cnn, resnet, lstm & rnn,
attention layers will be added in the future, if there is any feature request.
"""


def visualize_graph(final_tensor, exit_after_vis=True):
    """
    Visualize a pytorch flow graph

    Args:
        final_tensor: The last output tensor of the flow graph
        exit_after_vis: Whether to exit the whole program
            after visualization.
    """
    g = make_dot(final_tensor)
    g.view(quiet_view=True, quiet=True)
    if exit_after_vis:
        exit(0)



from PIL import Image
from typing import List
from machin.parallel import get_context
import numpy as np
import moviepy.editor as mpy
import matplotlib.pyplot as plt


def show_image(image: np.array,
               show_normalized: bool = True,
               pause_time: float = 0.01,
               title: str = ""):
    """
    Use matplotlib to show a single image. You may repeatedly call this method
    with the same ``title`` argument to show a video or a dynamically changing
    image.

    Args:
        image: A numpy array of shape (H, W, C) or (H, W), and with ``dtype``
            = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        show_normalized: Show normalized image alongside the original one.
        pause_time: Pause time between displaying current image and the next
            one.
        title: Title of the display window.
    """
    if np.issubdtype(image.dtype, np.integer):
        image = image.astype(np.floating) / 255

    fig = plt.figure(title, clear=True)
    fig.canvas.set_window_title(title)
    if show_normalized:
        ax = fig.add_subplot("121")
        ax.set_facecolor((0.0, 0.0, 0.0))
        ax.imshow(image, vmin=np.min(image), vmax=np.max(image))

        ax2 = fig.add_subplot("122")
        ax2.set_facecolor((0.0, 0.0, 0.0))
        pix_range = np.max(image) - np.min(image)
        if pix_range == 0:
            pix_range += 0.0001
        ax2.imshow((image - np.min(image)) / pix_range, vmin=0, vmax=1)
        plt.pause(pause_time)

    else:
        ax = fig.add_subplot("111")
        ax.set_facecolor((0.0, 0.0, 0.0))
        ax.imshow(image, vmin=np.min(image), vmax=np.max(image))
        plt.pause(pause_time)


def create_video(frames: List[np.array],
                 path: str,
                 extension: str = ".gif",
                 fps: int = 15):
    """
    Args:
        frames: A list of numpy arrays of shape (H, W, C) or (H, W), and with
            ``dtype`` = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        path: Path to save the video, without extension.
        extension: File extension.
        fps: frames per second.
    """
    for f in range(len(frames)):
        if np.issubdtype(frames[f].dtype, np.integer):
            frames[f] = frames[f].astype(np.uint8)
        elif np.issubdtype(frames[f].dtype, np.floating):
            frames[f] = (frames[f] * 255).astype(np.uint8)
        if frames[f].ndim == 2:
            # consider as a grey scale image
            frames[f] = np.stack([f, f, f], axis=-1)

    clip = mpy.ImageSequenceClip(frames, fps=fps)
    if extension.lower() == ".gif":
        clip.write_gif(path + extension, fps=fps, verbose=False, logger=None)
    else:
        clip.write_videofile(path + extension, fps=fps, verbose=False,
                             logger=None)


def create_video_subproc(frames: List[np.array],
                         path: str,
                         extension: str = ".gif",
                         fps=15,
                         daemon=True):
    """
    Create video with a subprocess, since it takes a lot of time for ``moviepy``
    to encode the video file.

    See Also:
         :func:`.create_video`

    Note:
        if ``daemon`` is true, then this function cannot be used in a
        daemonic subprocess.

    Args:
        frames: A list of numpy arrays of shape (H, W, C) or (H, W), and with
            ``dtype`` = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        path: Path to save the video, without extension.
        extension: File extension.
        fps: frames per second.
        daemon: Whether launching the saving process as a daemonic process.
    """
    if len(frames) == 0:
        raise RuntimeWarning("Empty frames sequence, file {} skipped"
                             .format(path + extension))
    p = get_context("spawn").Process(target=create_video,
                                     args=(frames, path, extension, fps))
    p.daemon = daemon
    p.start()


def create_image(image: np.array, path: str, extension=".png"):
    """
    Args:
        image: A numpy array of shape (H, W, C) or (H, W), and with
            ``dtype`` = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        path: Path to save the image, without extension.
        extension: Image extension.
    """
    if np.issubdtype(image.dtype, np.integer):
        image = image.astype(np.uint8)
    elif np.issubdtype(image.dtype, np.floating):
        image = (image * 255).astype(np.uint8)
    if image.sndim == 2:
        # consider as a grey scale image
        image = np.stack([image, image, image], axis=-1)
    image = Image.fromarray(image)
    image.save(path + extension)


def create_image_subproc(image: np.array, path, extension=".png", daemon=True):
    """
    Create image with a subprocess.

    See Also:
         :func:`.create_image`

    Note:
        if ``daemon`` is true, then this function cannot be used in a
        daemonic subprocess.

    Args:
        image: A numpy array of shape (H, W, C) or (H, W), and with
            ``dtype`` = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        path: Path to save the image, without extension.
        extension: Image extension.
        daemon: Whether launching the saving process as a daemonic process.
    """
    p = get_context("spawn").Process(target=create_image,
                                     args=(image, path, extension))
    p.daemon = daemon
    p.start()



from machin.utils.logging import default_logger
# pylint: disable=wildcard-import, unused-wildcard-import
from .ddpg import *


class TD3(DDPG):
    """
    TD3 framework. Which adds a additional pair of critic and target critic
    network to DDPG.
    """

    _is_top = ["actor", "critic", "critic2",
               "actor_target", "critic_target", "critic2_target"]
    _is_restorable = ["actor_target", "critic_target", "critic2_target"]

    def __init__(self,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 actor_target: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 critic_target: Union[NeuralNetworkModule, nn.Module],
                 critic2: Union[NeuralNetworkModule, nn.Module],
                 critic2_target: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 *_,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict, Dict] = (),
                 batch_size: int = 100,
                 update_rate: float = 0.005,
                 learning_rate: float = 0.001,
                 discount: float = 0.99,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 policy_noise_func: Callable = None,
                 reward_func: Callable = None,
                 action_trans_func: Callable = None,
                 visualize: bool = False,
                 **__):
        """
        See Also:
            :class:`.DDPG`

        Args:
            actor: Actor network module.
            actor_target: Target actor network module.
            critic: Critic network module.
            critic_target: Target critic network module.
            critic2: The second critic network module.
            critic2_target: The second target critic network module.
            optimizer: Optimizer used to optimize ``actor``, ``critic``,
            criterion: Criterion used to evaluate the value loss.
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`

            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            reward_func: Reward function used in training.
            action_trans_func: Action transform function, used to transform
                the raw output of your actor, by default it is:
                ``lambda act: {"action": act}``
            visualize: Whether visualize the network flow in the first pass.
        """
        super(TD3, self).__init__(
            actor, actor_target, critic, critic_target, optimizer, criterion,
            lr_scheduler=lr_scheduler,
            lr_scheduler_args=lr_scheduler_args[:2],
            lr_scheduler_kwargs=lr_scheduler_kwargs[:2],
            batch_size=batch_size,
            update_rate=update_rate,
            learning_rate=learning_rate,
            discount=discount,
            replay_size=replay_size,
            replay_device=replay_device,
            replay_buffer=replay_buffer,
            reward_func=reward_func,
            action_trans_func=action_trans_func,
            visualize=visualize
        )
        self.critic2 = critic2
        self.critic2_target = critic2_target
        self.critic2_optim = optimizer(self.critic2.parameters(),
                                       lr=learning_rate)

        # Make sure target and online networks have the same weight
        with t.no_grad():
            hard_update(self.critic2, self.critic2_target)

        if lr_scheduler is not None:
            self.critic2_lr_sch = lr_scheduler(
                self.critic2_optim,
                *lr_scheduler_args[2],
                **lr_scheduler_kwargs[2]
            )

        if policy_noise_func is None:
            default_logger.warning("Policy noise function is None, "
                                   "no policy noise will be applied "
                                   "during update!")
        self.policy_noise_func = (TD3._policy_noise_function
                                  if policy_noise_func is None
                                  else policy_noise_func)

    def criticize2(self,
                   state: Dict[str, Any],
                   action: Dict[str, Any],
                   use_target=False,
                   **__):
        """
        Use the second critic network to evaluate current value.

        Args:
            state: Current state.
            action: Current action.
            use_target: Whether to use the target network.

        Returns:
            Value of shape ``[batch_size, 1]``.
        """
        if use_target:
            return safe_call(self.critic2_target, state, action)
        else:
            return safe_call(self.critic2, state, action)

    def update(self,
               update_value=True,
               update_policy=True,
               update_target=True,
               concatenate_samples=True,
               **__):
        # DOC INHERITED
        batch_size, (state, action, reward, next_state, terminal, *others) = \
            self.replay_buffer.sample_batch(self.batch_size,
                                            concatenate_samples,
                                            sample_attrs=[
                                                "state", "action",
                                                "reward", "next_state",
                                                "terminal", "*"
                                            ])

        # Update critic network first.
        # Generate value reference :math: `y_i` using target actor and
        # target critic.
        with t.no_grad():
            next_action = self.action_transform_func(
                self.policy_noise_func(
                    self.act(next_state, True)
                ),
                next_state, *others
            )
            next_value = self.criticize(next_state, next_action, True)
            next_value2 = self.criticize2(next_state, next_action, True)
            next_value = t.min(next_value, next_value2)
            next_value = next_value.view(batch_size, -1)
            y_i = self.reward_func(reward, self.discount, next_value, terminal,
                                   *others)

        cur_value = self.criticize(state, action)
        cur_value2 = self.criticize2(state, action)
        value_loss = self.criterion(cur_value, y_i.to(cur_value.device))
        value_loss2 = self.criterion(cur_value2, y_i.to(cur_value.device))

        if self.visualize:
            self.visualize_model(value_loss, "critic")

        if update_value:
            self.critic.zero_grad()
            value_loss.backward()
            self.critic_optim.step()
            self.critic2.zero_grad()
            value_loss2.backward()
            self.critic2_optim.step()

        # Update actor network
        cur_action = self.action_transform_func(self.act(state), state, *others)
        act_value = self.criticize(state, cur_action)

        # "-" is applied because we want to maximize J_b(u),
        # but optimizer workers by minimizing the target
        act_policy_loss = -act_value.mean()

        if self.visualize:
            self.visualize_model(act_policy_loss, "actor")

        if update_policy:
            self.actor.zero_grad()
            act_policy_loss.backward()
            self.actor_optim.step()

        # Update target networks
        if update_target:
            soft_update(self.actor_target, self.actor, self.update_rate)
            soft_update(self.critic_target, self.critic, self.update_rate)
            soft_update(self.critic2_target, self.critic2, self.update_rate)

        # use .item() to prevent memory leakage
        return (-act_policy_loss.item(),
                (value_loss.item() + value_loss2.item()) / 2)

    @staticmethod
    def _policy_noise_function(actions, *_):
        return actions

    def load(self, model_dir: str, network_map: Dict[str, str] = None,
             version: int = -1):
        # DOC INHERITED
        TorchFramework.load(self, model_dir, network_map, version)
        with t.no_grad():
            hard_update(self.actor, self.actor_target)
            hard_update(self.critic, self.critic_target)
            hard_update(self.critic2, self.critic2_target)



import copy
import itertools
import numpy as np
from machin.frame.transition import TransitionBase
from machin.utils.logging import default_logger
from machin.parallel.pool import Pool
from machin.parallel.assigner import ModelAssigner
# pylint: disable=wildcard-import, unused-wildcard-import
from .ddpg import *


def _average_parameters(*params: t.Tensor):
    # Average parameters from all sub policies, store the
    # averaged result in the first sub policy, then broadcast
    # result to remaining sub policies.
    target_param = params[0]
    target_param.data.copy_(
        t.mean(
            t.stack([p.to(target_param.device) for p in params], dim=0),
            dim=0
        )
    )
    for param in params[1:]:
        param.data.copy_(target_param)


class MultiAgentTransition(TransitionBase):
    """
    Multi agent Transition class used in MADDPG.
    """
    state = None  # type: Dict[str, t.Tensor]
    all_states = None  # type: Dict[str, t.Tensor]
    all_actions = None  # type: Dict[str, t.Tensor]
    all_next_states = None  # type: Dict[str, t.Tensor]
    reward = None  # type: Union[float, t.Tensor]
    terminal = None  # type: bool

    def __init__(self,
                 state: Dict[str, t.Tensor],
                 all_states: Dict[str, t.Tensor],
                 all_actions: Dict[str, t.Tensor],
                 all_next_states: Dict[str, t.Tensor],
                 reward: Union[float, t.Tensor],
                 index: int,
                 terminal: bool,
                 **kwargs):
        super(MultiAgentTransition, self).__init__(
            major_attr=["state", "all_states", "all_actions",
                        "all_next_states"],
            sub_attr=["reward", "terminal"],
            custom_attr=kwargs.keys(),
            major_data=[state, all_states, all_actions, all_next_states],
            sub_data=[reward, terminal, index],
            custom_data=kwargs.values()
        )


class MADDPG(TorchFramework):
    """
    MADDPG is a centralized multi-agent training framework, it alleviates the
    unstable reward problem caused by the disturbance of other agents by
    gathering all agents observations and train a global critic. This global
    critic observes all actions and all states from all agents.
    """
    # Since the number of sub-policies is automatically determined,
    # they are not considered here.
    _is_top = ["actor_target", "critic_target"]
    _is_restorable = ["actor_target", "critic_target"]

    def __init__(self,
                 agent_num,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 actor_target: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 critic_target: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 available_devices: Union[list, None] = None,
                 sub_policy_num: int = 1,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict, Dict] = (),
                 batch_size: int = 100,
                 update_rate: float = 0.005,
                 learning_rate: float = 0.001,
                 discount: float = 0.99,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 reward_func: Callable = None,
                 action_concat_func: Callable = None,
                 action_alter_func: Callable = None,
                 state_split_func: Callable = None,
                 visualize: bool = False,
                 pool: Any = None):
        """
        See Also:
            :class:`.DDPG`

        Hint:
            Please reference:

            - :meth:`MADDPG.action_concat_function`
            - :meth:`MADDPG.action_alter_function`
            - :meth:`MADDPG.state_split_function`

            for ``action_concat_func``, ``action_alter_func``, and
            ``state_split_func`` design, if your actor does not output
            a simple action of shape ``[batch_size, action_dim]`` which
            should be directly accepted by your critic.

        Args:
            actor: Actor network module.
            actor_target: Target actor network module.
            critic: Critic network module.
            critic_target: Target critic network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            criterion: Criterion used to evaluate the value loss.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:
                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            reward_func: Reward function used in training.
            action_concat_func: Action concatenation function.
            action_alter_func: Action alternation function.
            state_split_func: All states spli function.
            visualize: Whether visualize the network flow in the first pass.
        """
        self.batch_size = batch_size
        self.update_rate = update_rate
        self.discount = discount
        self.visualize = visualize
        self.agent_num = agent_num

        self.actors = [copy.deepcopy(actor)
                       for _ in range(sub_policy_num)]
        self.actor_targets = [copy.deepcopy(actor_target)
                              for _ in range(sub_policy_num)]
        self.critics = [copy.deepcopy(critic)
                        for _ in range(sub_policy_num)]
        self.critic_targets = [copy.deepcopy(critic_target)
                               for _ in range(sub_policy_num)]
        self.actor_optims = [optimizer(ac.parameters(), lr=learning_rate)
                             for ac in self.actors]
        self.critic_optims = [optimizer(cr.parameters(), lr=learning_rate)
                              for cr in self.critics]
        self.sub_policy_num = sub_policy_num
        self.pool = Pool() if pool is None else pool
        self.replay_buffer = (Buffer(replay_size, replay_device)
                              if replay_buffer is None
                              else replay_buffer)

        # Distribute models to available devices.
        if available_devices is not None and available_devices:
            nets = self.actors + self.actor_targets + \
                   self.critics + self.critic_targets
            # Only actors and critics are related
            connections = {(i, i + self.sub_policy_num * 2): 1
                           for i in range(self.sub_policy_num)}
            # We do not need the assigner after construction.
            _assigner = ModelAssigner(nets, connections, available_devices)

            # For debugging:
            (act_assign, act_target_assign,
             critic_assign, critic_target_assign) = \
                np.array_split(_assigner.assignment, 4)
            default_logger.log("Actors assigned to:")
            default_logger.log(act_assign)
            default_logger.log("Actors (target) assigned to:")
            default_logger.log(act_target_assign)
            default_logger.log("Critics assigned to:")
            default_logger.log(critic_assign)
            default_logger.log("Critics (target) assigned to:")
            default_logger.log(critic_target_assign)

        # Create wrapper for target actors and target critics.
        # So their parameters can be saved.
        self.actor_target = nn.Module()
        self.critic_target = nn.Module()
        for actor_t, idx in zip(self.actor_targets,
                                range(self.sub_policy_num)):
            self.actor_target.add_module("actor_{}".format(idx), actor_t)

        for critic_t, idx in zip(self.critic_targets,
                                 range(self.sub_policy_num)):
            self.critic_target.add_module("critic_{}".format(idx), critic_t)

        # Make sure target and online networks have the same weight
        with t.no_grad():
            self.pool.starmap(hard_update,
                              zip(self.actors, self.actor_targets))
            self.pool.starmap(hard_update,
                              zip(self.critics, self.critic_targets))

        if lr_scheduler is not None:
            self.actor_lr_schs = [lr_scheduler(ac_opt,
                                               *lr_scheduler_args[0],
                                               *lr_scheduler_kwargs[0])
                                  for ac_opt in self.actor_optims]
            self.critic_lr_schs = [lr_scheduler(cr_opt,
                                                *lr_scheduler_args[1],
                                                *lr_scheduler_kwargs[1])
                                   for cr_opt in self.critic_optims]

        self.criterion = criterion

        self.reward_func = (MADDPG.bellman_function
                            if reward_func is None
                            else reward_func)

        self.action_alter_func = (MADDPG.action_alter_function
                                  if action_alter_func is None
                                  else action_alter_func)

        self.action_concat_func = (MADDPG.action_concat_function
                                   if action_concat_func is None
                                   else action_concat_func)

        self.state_split_func = (MADDPG.state_split_function
                                 if state_split_func is None
                                 else state_split_func)

        super(MADDPG, self).__init__()

    def act(self,
            state: Dict[str, Any],
            use_target: bool = False,
            index: int = -1,
            **__):
        """
        Use actor network to produce an action for the current state.

        Args:
            state: Current state.
            use_target: Whether use the target network.
            index: The sub-policy index to use.

        Returns:
            Action of shape ``[batch_size, action_dim]``.
        """
        if index not in range(self.sub_policy_num):
            index = np.random.randint(0, self.sub_policy_num)

        if use_target:
            return safe_call(self.actor_targets[index], state)
        else:
            return safe_call(self.actors[index], state)

    def act_with_noise(self,
                       state: Dict[str, Any],
                       noise_param: Tuple = (0.0, 1.0),
                       ratio: float = 1.0,
                       mode: str = "uniform",
                       use_target: bool = False,
                       index: int = -1,
                       **__):
        """
        Use actor network to produce a noisy action for the current state.

        See Also:
             :mod:`machin.frame.noise.action_space_noise`

        Args:
            state: Current state.
            noise_param: Noise params.
            ratio: Noise ratio.
            mode: Noise mode. Supported are:
                ``"uniform", "normal", "clipped_normal", "ou"``
            use_target: Whether use the target network.
            index: The sub-policy index to use.

        Returns:
            Noisy action of shape ``[batch_size, action_dim]``.
        """
        if mode == "uniform":
            return add_uniform_noise_to_action(
                self.act(state, use_target, index), noise_param, ratio
            )
        if mode == "normal":
            return add_normal_noise_to_action(
                self.act(state, use_target, index), noise_param, ratio
            )
        if mode == "clipped_normal":
            return add_clipped_normal_noise_to_action(
                self.act(state, use_target, index), noise_param, ratio
            )
        if mode == "ou":
            return add_ou_noise_to_action(
                self.act(state, use_target, index), noise_param, ratio
            )
        raise RuntimeError("Unknown noise type: " + str(mode))

    def act_discreet(self,
                     state: Dict[str, Any],
                     use_target: bool = False,
                     index: int = -1):
        """
        Use actor network to produce a discreet action for the current state.

        Notes:
            actor network must output a probability tensor, of shape
            (batch_size, action_dims), and has a sum of 1 for each row
            in dimension 1.

        Args:
            state: Current state.
            use_target: Whether to use the target network.
            index: The sub-policy index to use.

        Returns:
            Action of shape ``[batch_size, 1]``.
        """
        if index not in range(self.sub_policy_num):
            index = np.random.randint(0, self.sub_policy_num)

        if use_target:
            result = safe_call(self.actor_targets[index], state)
        else:
            result = safe_call(self.actors[index], state)

        assert_output_is_probs(result)
        batch_size = result.shape[0]
        result = t.argmax(result, dim=1).view(batch_size, 1)
        return result

    def act_discreet_with_noise(self,
                                state: Dict[str, Any],
                                use_target: bool = False,
                                index=-1):
        """
        Use actor network to produce a noisy discreet action for
        the current state.

        Notes:
            actor network must output a probability tensor, of shape
            (batch_size, action_dims), and has a sum of 1 for each row
            in dimension 1.

        Args:
            state: Current state.
            use_target: Whether to use the target network.
            index: The sub-policy index to use.

        Returns:
            Noisy action of shape ``[batch_size, 1]``.
        """
        if index not in range(self.sub_policy_num):
            index = np.random.randint(0, self.sub_policy_num)

        if use_target:
            result = safe_call(self.actor_targets[index], state)
        else:
            result = safe_call(self.actors[index], state)

        assert_output_is_probs(result)
        dist = Categorical(result)
        batch_size = result.shape[0]
        return dist.sample([batch_size, 1])

    def criticize(self, all_states, all_actions, use_target=False, index=-1):
        """
        Use critic network to evaluate current value.

        Args:
            all_states: Current states of all actors.
            all_actions: Current actions of all actors.
            use_target: Whether to use the target network.
            index: The sub-critic index to use.

        Returns:
            Value of shape ``[batch_size, 1]``.
        """
        if index not in range(self.sub_policy_num):
            index = np.random.randint(0, self.sub_policy_num)

        if use_target:
            return safe_call(self.critic_targets[index],
                             all_states, all_actions)
        else:
            return safe_call(self.critics[index],
                             all_states, all_actions)

    def store_transition(self, transition: Union[MultiAgentTransition, Dict]):
        """
        Add a transition sample to the replay buffer.
        """
        self.replay_buffer.append(transition, required_attrs=(
            "state", "all_states", "all_actions", "all_next_states",
            "reward", "terminal", "index"
        ))

    def store_episode(self, episode: List[Union[MultiAgentTransition, Dict]]):
        """
        Add a full episode of transition samples to the replay buffer.
        """
        for trans in episode:
            self.replay_buffer.append(trans, required_attrs=(
                "state", "all_states", "all_actions", "all_next_states",
                "reward", "terminal", "index"
            ))

    def update(self,
               update_value=True,
               update_policy=True,
               update_target=True,
               concatenate_samples=True,
               average_target_parameter=False):
        """
        Update network weights by sampling from replay buffer.

        Args:
            update_value: Whether to update the Q network.
            update_policy: Whether to update the actor network.
            update_target: Whether to update targets.
            concatenate_samples: Whether to concatenate the samples.
            average_target_parameter: Whether to average sub target networks,
                including actors and critics.
        Returns:
            mean value of estimated policy value, value loss
        """
        batch_size, (state, all_states, all_actions, all_next_states,
                     reward, terminal, agent_indexes, *others) = \
            self.replay_buffer.sample_batch(self.batch_size,
                                            concatenate_samples,
                                            sample_attrs=[
                                                "state", "all_states",
                                                "all_actions",
                                                "all_next_states",
                                                "reward", "terminal", "index",
                                                "*"
                                            ])

        def update_inner(i):
            with t.no_grad():
                # Produce all_next_actions for all_next_states, using target i,
                # so the target critic can evaluate the value of the next step.
                all_next_actions_t = \
                    self.action_concat_func(
                        self.act(
                            self.state_split_func(
                                all_next_states, batch_size,
                                self.agent_num, others
                            ),
                            True, i
                        ),
                        batch_size, self.agent_num, others
                    )

            # Update critic network first
            # Generate target value using target critic.
            with t.no_grad():
                next_value = self.criticize(all_next_states,
                                            all_next_actions_t,
                                            True, i)
                next_value = next_value.view(batch_size, -1)
                y_i = self.reward_func(reward, self.discount, next_value,
                                       terminal, others)

            # action contain actions of all agents, same for state
            cur_value = self.criticize(all_states, all_actions, index=i)
            value_loss = self.criterion(cur_value, y_i.to(cur_value.device))

            if update_value:
                self.critics[i].zero_grad()
                value_loss.backward()
                self.critic_optims[i].step()

            # Update actor network
            cur_all_actions = copy.deepcopy(all_actions)
            cur_all_actions = self.action_alter_func(
                self.act(state, index=i), cur_all_actions, agent_indexes,
                batch_size, self.agent_num, others
            )
            act_value = self.criticize(state, cur_all_actions, index=i)

            # "-" is applied because we want to maximize J_b(u),
            # but optimizer workers by minimizing the target
            act_policy_loss = -act_value.mean()

            if update_policy:
                self.actors[i].zero_grad()
                act_policy_loss.backward()
                self.actor_optims[i].step()

            # Update target networks
            if update_target:
                soft_update(self.actor_targets[i], self.actors[i],
                            self.update_rate)
                soft_update(self.critic_targets[i], self.critics[i],
                            self.update_rate)

            return act_policy_loss.item(), value_loss.item()

        all_loss = self.pool.map(update_inner, range(self.sub_policy_num))
        mean_loss = t.tensor(all_loss).mean(dim=0)

        if average_target_parameter:
            self.average_target_parameters()

        # returns action value and policy loss
        return -mean_loss[0].item(), mean_loss[1].item()

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "actor_lr_schs"):
            for actor_lr_sch in self.actor_lr_schs:
                actor_lr_sch.step()
        if hasattr(self, "critic_lr_schs"):
            for critic_lr_sch in self.critic_lr_schs:
                critic_lr_sch.step()

    def load(self, model_dir, network_map=None, version=-1):
        # DOC INHERITED
        super(MADDPG, self).load(model_dir, network_map, version)
        with t.no_grad():
            self.pool.starmap(hard_update,
                              zip(self.actors, self.actor_targets))
            self.pool.starmap(hard_update,
                              zip(self.critics, self.critic_targets))

    def average_target_parameters(self):
        """
        Average parameters of sub-policies and sub-critics. Averaging
        is performed on target networks.
        """
        with t.no_grad():
            actor_params = [net.parameters() for net in self.actor_targets]
            critic_params = [net.parameters() for net in self.critic_targets]
            self.pool.starmap(
                _average_parameters,
                itertools.chain(zip(*actor_params), zip(*critic_params))
            )

    @staticmethod
    def action_alter_function(raw_output_action, all_actions, indexes,
                              batch_size, agent_num, *_):
        """
        This function is used to alternate an action inside all actions,
        using output from the online actor network.

        Args:
            raw_output_action: Raw output of actor.
            all_actions: All actions of all agents.
            indexes: Agent index among all agents.
            batch_size: Sampled batch size.
            agent_num: Number of agents.

        Returns:
            Alternated all actions.
        """
        all_actions["action"][indexes] = \
            raw_output_action.view(batch_size, agent_num, -1)
        return all_actions

    @staticmethod
    def state_split_function(all_states: Dict[str, t.Tensor],
                             batch_size: int,
                             agent_num: int,
                             *_):
        """
        This function is used to split states from multiple agents into
        batched single states, usable by actor network.

        Args:
            all_states: All states of all agents.
            batch_size: Sampled batch size.
            agent_num: Number of agents.
        Returns:
            Splitted states.
        """
        all_states["state"] = all_states["state"]\
            .view(batch_size * agent_num, -1)
        return all_states

    @staticmethod
    def action_concat_function(raw_output_action, batch_size, agent_num, *_):
        """
        This function is used to transform the actions produced by actor
        from the splitted states, to the final output of all actions of
        all agents.

        Args:
            raw_output_action: Raw output of actor.
            batch_size: Sampled batch size.
            agent_num: Number of agents.

        Returns:
            Concatenated actions.
        """
        return {"action": raw_output_action.view(batch_size, agent_num)}

    @staticmethod
    def bellman_function(reward, discount, next_value, terminal, *_):
        next_value = next_value.to(reward.device)
        terminal = terminal.to(reward.device)
        return reward + discount * (1 - terminal) * next_value



def disable_view_window():
    # Disable pop up windows and render in background
    # by injecting custom viewer constructor.
    from gym.envs.classic_control import rendering
    org_constructor = rendering.Viewer.__init__

    def constructor(self, *args, **kwargs):
        org_constructor(self, *args, **kwargs)
        self.window.set_visible(visible=False)

    rendering.Viewer.__init__ = constructor



import os
import time
import torch as t
from datetime import timedelta
from torch.multiprocessing import Process
import torch.distributed as dist

rank = None
messages = 1000
world_size = 4
storage = t.zeros([1000])
storage1 = t.zeros([1000])

def run(i):
    global rank
    rank = i
    dist.init_process_group(backend="gloo",
                            init_method="env://",
                            timeout=timedelta(seconds=2.5),
                            world_size=world_size,
                            rank=rank)

    if i == 0:
        time.sleep(2000)
        print("Process 0 exit.")
        exit(-1)
    t = time.time()
    for j in range(messages):
        for r in range(world_size):
            print("round {}: {}, {} begin".format(j, rank, r))
            try:
                dist.isend(storage, r)
                #dist.recv(storage1)
            except RuntimeError:
                print("round {}: {}, {} error".format(j, rank, r))
            else:
                print("round {}: {}, {}, {:.5f}".format(j, rank, r, time.time() - t))
    print(time.time() - t)
    dist.destroy_process_group()


if __name__ == "__main__":
    os.environ['MASTER_ADDR'] = "localhost"
    os.environ['MASTER_PORT'] = "29530"
    ps = []
    for i in range(world_size):
        p = Process(target=run, args=(i,))
        p.start()
        ps.append(p)

    for p in ps:
        p.join()



from .buffer import Buffer
from .buffer_d import DistributedBuffer
from .prioritized_buffer import PrioritizedBuffer
from .prioritized_buffer_d import DistributedPrioritizedBuffer

__all__ = ["Buffer", "DistributedBuffer",
           "PrioritizedBuffer", "DistributedPrioritizedBuffer"]



from .dqn_per import *
from .ddpg_per import *
from ..buffers.prioritized_buffer_d import DistributedPrioritizedBuffer
from machin.parallel.server import PushPullModelServer
from machin.parallel.distributed import RpcGroup


class DQNApex(DQNPer):
    """
    Massively parallel version of a Double DQN with prioritized replay.

    The pull function is invoked before using ``act_discreet``,
    ``act_discreet_with_noise`` and ``criticize``.

    The push function is invoked after ``update``.
    """
    def __init__(self,
                 qnet: Union[NeuralNetworkModule, nn.Module],
                 qnet_target: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 worker_group: RpcGroup,
                 trainer_group: RpcGroup,
                 *_,
                 pull_function: Callable = None,
                 push_function: Callable = None,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
                 batch_size: int = 100,
                 update_rate: float = 0.005,
                 learning_rate: float = 0.001,
                 discount: float = 0.99,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 reward_func: Callable = None,
                 visualize: bool = False,
                 **__):
        """
        See Also:
            :class:`.DQNPer`

        Hint:
            Your push and pull function will be called like::

                function(actor_model, "actor")

            The default implementation of pull and push functions
            is provided by :class:`.PushPullModelServer`.

        Args:
            qnet: Q network module.
            qnet_target: Target Q network module.
            optimizer: Optimizer used to optimize ``qnet``.
            criterion: Criterion used to evaluate the value loss.
            worker_group: Rpc group of roll out workers.
            trainer_group: Rpc group of model trainers.
            pull_function: User defined function used to pull the newest model.
            push_function: User defined function used to push the newest model.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            reward_func: Reward function used in training.
            visualize: Whether visualize the network flow in the first pass.
        """
        super(DQNApex, self).__init__(qnet, qnet_target, optimizer, criterion,
                                      lr_scheduler=lr_scheduler,
                                      lr_scheduler_args=lr_scheduler_args,
                                      lr_scheduler_kwargs=lr_scheduler_kwargs,
                                      batch_size=batch_size,
                                      update_rate=update_rate,
                                      learning_rate=learning_rate,
                                      discount=discount,
                                      replay_size=replay_size,
                                      replay_device=replay_device,
                                      replay_buffer=replay_buffer,
                                      reward_func=reward_func,
                                      visualize=visualize)

        # Currently, worker group is not used, reserved.
        self.worker_group = worker_group
        self.trainer_group = trainer_group

        self.replay_buffer = DistributedPrioritizedBuffer(
            buffer_size=replay_size, buffer_group=trainer_group
        )

        if push_function is None or pull_function is None:
            self.pp = PushPullModelServer(trainer_group)
            self.pull_function = self.pp.pull
            self.push_function = self.pp.push
        else:
            self.pull_function = pull_function
            self.push_function = push_function

    def act_discreet(self,
                     state: Dict[str, Any],
                     use_target: bool = False,
                     **__):
        # DOC INHERITED
        if use_target:
            self.pull_function(self.qnet_target, "qnet_target")
        else:
            self.pull_function(self.qnet, "qnet")
        return super(DQNApex, self).act_discreet(state, use_target)

    def act_discreet_with_noise(self,
                                state: Dict[str, Any],
                                use_target: bool = False,
                                **__):
        # DOC INHERITED
        if use_target:
            self.pull_function(self.qnet_target, "qnet_target")
        else:
            self.pull_function(self.qnet, "qnet")
        return super(DQNApex, self).act_discreet_with_noise(state, use_target)

    def criticize(self,
                  state: Dict[str, Any],
                  use_target: bool = False,
                  **__):
        # DOC INHERITED
        if use_target:
            self.pull_function(self.qnet_target, "qnet_target")
        else:
            self.pull_function(self.qnet, "qnet")
        return super(DQNApex, self).criticize(state, use_target)

    def update(self,
               update_value=True,
               update_target=True,
               concatenate_samples=True,
               **__):
        # DOC INHERITED
        result = super(DQNApex, self).update(update_value, update_target,
                                             concatenate_samples)
        if update_target:
            self.push_function(self.qnet_target, "qnet_target")
        if update_value:
            self.push_function(self.qnet, "qnet")
        return result


class DDPGApex(DDPGPer):
    """
    Massively parallel version of a DDPG with prioritized replay.

    The pull function is invoked before using
    ``act``, ``act_with_noise``, ``act_discreet``,
    ``act_discreet_with_noise`` and ``criticize``.

    The push function is invoked after ``update``.
    """
    def __init__(self,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 actor_target: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 critic_target: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 worker_group: RpcGroup,
                 trainer_group: RpcGroup,
                 *_,
                 pull_function: Callable = None,
                 push_function: Callable = None,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
                 batch_size: int = 100,
                 update_rate: float = 0.005,
                 learning_rate: float = 0.001,
                 discount: float = 0.99,
                 replay_size: int = 500000,
                 reward_func: Callable = None,
                 action_trans_func: Callable = None,
                 visualize: bool = False,
                 **__):
        """
        See Also:
            :class:`.DDPGPer`

        Hint:
            Your push and pull function will be called like::

                function(actor_model, "actor")

            The default implementation of pull and push functions
            is provided by :class:`.PushPullModelServer`.

        Args:
            actor: Actor network module.
            actor_target: Target actor network module.
            critic: Critic network module.
            critic_target: Target critic network module.
            optimizer: Optimizer used to optimize ``qnet``.
            criterion: Criterion used to evaluate the value loss.
            worker_group: Rpc group of roll out workers.
            trainer_group: Rpc group of model trainers.
            pull_function: User defined function used to pull the newest model.
            push_function: User defined function used to push the newest model.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Size of the local replay buffer.
            reward_func: Reward function used in training.
            visualize: Whether visualize the network flow in the first pass.
        """
        super(DDPGApex, self).__init__(actor, actor_target,
                                       critic, critic_target,
                                       optimizer, criterion,
                                       lr_scheduler=lr_scheduler,
                                       lr_scheduler_args=lr_scheduler_args,
                                       lr_scheduler_kwargs=lr_scheduler_kwargs,
                                       batch_size=batch_size,
                                       update_rate=update_rate,
                                       learning_rate=learning_rate,
                                       discount=discount,
                                       replay_size=0,
                                       replay_device="cpu",
                                       replay_buffer=None,
                                       reward_func=reward_func,
                                       action_trans_func=action_trans_func,
                                       visualize=visualize)

        # Currently, worker group is not used, reserved.
        self.worker_group = worker_group
        self.trainer_group = trainer_group

        self.replay_buffer = DistributedPrioritizedBuffer(
            buffer_size=replay_size, buffer_group=trainer_group
        )

        if push_function is None or pull_function is None:
            self.pp = PushPullModelServer(trainer_group)
            self.pull_function = self.pp.pull
            self.push_function = self.pp.push
        else:
            self.pull_function = pull_function
            self.push_function = push_function

    def act(self,
            state: Dict[str, Any],
            use_target: bool = False,
            **__):
        # DOC INHERITED
        if use_target:
            self.pull_function(self.actor_target, "actor_target")
        else:
            self.pull_function(self.actor, "actor")
        return super(DDPGApex, self).act(state, use_target)

    def act_with_noise(self,
                       state: Dict[str, Any],
                       noise_param: Tuple = (0.0, 1.0),
                       ratio: float = 1.0,
                       mode: str = "uniform",
                       use_target: bool = False,
                       **__):
        # DOC INHERITED
        if use_target:
            self.pull_function(self.actor_target, "actor_target")
        else:
            self.pull_function(self.actor, "actor")
        return super(DDPGApex, self).act_with_noise(state,
                                                    noise_param=noise_param,
                                                    ratio=ratio,
                                                    mode=mode,
                                                    use_target=use_target)

    def act_discreet(self,
                     state: Dict[str, Any],
                     use_target: bool = False,
                     **__):
        # DOC INHERITED
        if use_target:
            self.pull_function(self.actor_target, "actor_target")
        else:
            self.pull_function(self.actor, "actor")
        return super(DDPGApex, self).act_discreet(state, use_target)

    def act_discreet_with_noise(self,
                                state: Dict[str, Any],
                                use_target: bool = False,
                                **__):
        # DOC INHERITED
        if use_target:
            self.pull_function(self.actor_target, "actor_target")
        else:
            self.pull_function(self.actor, "actor")
        return super(DDPGApex, self).act_discreet_with_noise(state, use_target)

    def criticize(self,
                  state: Dict[str, Any],
                  action: Dict[str, Any],
                  use_target: bool = False,
                  **__):
        # DOC INHERITED
        if use_target:
            self.pull_function(self.critic_target, "critic_target")
        else:
            self.pull_function(self.critic, "critic")
        return super(DDPGApex, self).criticize(state, action, use_target)

    def update(self,
               update_value=True,
               update_policy=True,
               update_target=True,
               concatenate_samples=True,
               **__):
        # DOC INHERITED
        result = super(DDPGApex, self).update(update_value, update_policy,
                                              update_target,
                                              concatenate_samples)
        if update_target:
            self.push_function(self.critic_target, "critic_target")
            self.push_function(self.actor_target, "actor_target")
        if update_value:
            self.push_function(self.critic, "critic")
            self.push_function(self.actor, "actor")
        return result



# pylint: disable=wildcard-import, unused-wildcard-import
from .ddpg import *


class HDDPG(DDPG):
    """
    HDDPG framework.
    """
    def __init__(self,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 actor_target: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 critic_target: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 *_,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
                 batch_size: int = 100,
                 update_rate: float = 0.005,
                 learning_rate: float = 0.001,
                 discount: float = 0.99,
                 q_increase_rate: float = 1.0,
                 q_decrease_rate: float = 0.5,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 reward_func: Callable = None,
                 action_trans_func: Callable = None,
                 visualize: bool = False,
                 **__):
        """
        Args:
            actor: Actor network module.
            actor_target: Target actor network module.
            critic: Critic network module.
            critic_target: Target critic network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            criterion: Critierion used to evaluate the value loss.
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`

            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            reward_func: Reward function used in training.
            action_trans_func: Action transform function, used to transform
                the raw output of your actor, by default it is:
                ``lambda act: {"action": act}``
            visualize: Whether visualize the network flow in the first pass.
            q_increase_rate: The increase ratio multiplied to q value
                and target value difference when difference is positive.
            q_decrease_rate: The decrease ratio multiplied to q value
                and target value difference when difference is negative.

        See Also:
            :class:`.DDPG`
        """
        super(HDDPG, self).__init__(actor, actor_target, critic, critic_target,
                                    optimizer, criterion,
                                    lr_scheduler=lr_scheduler,
                                    lr_scheduler_args=lr_scheduler_args,
                                    lr_scheduler_kwargs=lr_scheduler_kwargs,
                                    batch_size=batch_size,
                                    update_rate=update_rate,
                                    learning_rate=learning_rate,
                                    discount=discount,
                                    replay_size=replay_size,
                                    replay_device=replay_device,
                                    replay_buffer=replay_buffer,
                                    reward_func=reward_func,
                                    action_trans_func=action_trans_func,
                                    visualize=visualize)
        self.q_increase_rate = q_increase_rate
        self.q_decrease_rate = q_decrease_rate

    def update(self,
               update_value=True,
               update_policy=True,
               update_target=True,
               concatenate_samples=True,
               **__):
        # DOC INHERITED
        batch_size, (state, action, reward, next_state, terminal, *others) = \
            self.replay_buffer.sample_batch(self.batch_size,
                                            concatenate_samples,
                                            sample_attrs=[
                                                "state", "action",
                                                "reward", "next_state",
                                                "terminal", "*"
                                            ])

        # Update critic network first.
        # Generate value reference :math: `y_i` using target actor and
        # target critic.
        with t.no_grad():
            next_action = self.action_transform_func(self.act(next_state, True),
                                                     next_state,
                                                     *others)
            next_value = self.criticize(next_state, next_action, True)
            next_value = next_value.view(batch_size, -1)
            y_i = self.reward_func(reward, self.discount, next_value,
                                   terminal, *others)

        cur_value = self.criticize(state, action)
        value_diff = y_i.to(cur_value.device) - cur_value
        value_change = t.where(value_diff > 0,
                               value_diff * self.q_increase_rate,
                               value_diff * self.q_decrease_rate)
        value_loss = self.criterion(cur_value,
                                    (cur_value + value_change).detach())

        if self.visualize:
            self.visualize_model(value_loss, "critic")

        if update_value:
            self.critic.zero_grad()
            value_loss.backward()
            self.critic_optim.step()

        # Update actor network
        cur_action = self.action_transform_func(self.act(state), state, *others)
        act_value = self.criticize(state, cur_action)

        # "-" is applied because we want to maximize J_b(u),
        # but optimizer workers by minimizing the target
        act_policy_loss = -act_value.mean()

        if self.visualize:
            self.visualize_model(act_policy_loss, "actor")

        if update_policy:
            self.actor.zero_grad()
            act_policy_loss.backward()
            self.actor_optim.step()

        # Update target networks
        if update_target:
            soft_update(self.actor_target, self.actor, self.update_rate)
            soft_update(self.critic_target, self.critic, self.update_rate)

        # use .item() to prevent memory leakage
        return -act_policy_loss.item(), value_loss.item()



from typing import Tuple, Iterable, Union
import torch as t

from .generator import OrnsteinUhlenbeckNoiseGen


DEFAULT_OU_GEN = None

NoiseParam = Union[Iterable[Tuple], Tuple]


def add_uniform_noise_to_action(action: t.Tensor,
                                noise_param: NoiseParam = (0.0, 1.0),
                                ratio: float = 1.0):
    """
    Add uniform noise to action tensor.

    Hint:
        The innermost tuple contains: ``(uniform_min, uniform_max)``

        If ``noise_param`` is ``Tuple[float, float]``, then the same uniform
        noise will be added to ``action[*, :]``.

        If ``noise_param`` is ``Iterable[Tuple[float, float]]``, then for each
        ``action[*, i]`` slice i, uniform noise with ``noise_param[i]`` will
        be added respectively.

    Args:
        action: Raw action.
        noise_param: Param of the uniform noise.
        ratio: Sampled noise is multiplied with this ratio.

    Returns:
        Action with uniform noise.
    """
    if isinstance(noise_param[0], tuple):
        if len(noise_param) != action.shape[-1]:
            raise ValueError("Noise param length doesn't match "
                             "the last dimension of action")
        noise = t.rand(action.shape, device=action.device)
        for i in range(action.shape[-1]):
            noi_p = noise_param[i]
            noise.view(-1, noise.shape[-1])[:, i] *= noi_p[1] - noi_p[0]
            noise.view(-1, noise.shape[-1])[:, i] += noi_p[0]
    else:
        noise_param = noise_param  # type: Tuple[float, float]
        noise = t.rand(action.shape, device=action.device) \
            * (noise_param[1] - noise_param[0]) + noise_param[0]
    return action + noise * ratio


def add_clipped_normal_noise_to_action(
        action: t.Tensor, noise_param: NoiseParam = (0.0, 1.0, -1.0, 1.0),
        ratio=1.0):
    """
    Add clipped normal noise to action tensor.

    Hint:
        The innermost tuple contains:
        ``(normal_mean, normal_sigma, clip_min, clip_max)``

        If ``noise_param`` is ``Tuple[float, float, float, float]``,
        then the same clipped normal noise will be added to ``action[*, :]``.

        If ``noise_param`` is ``Iterable[Tuple[float, float, float, float]]``,
        then for each ``action[*, i]`` slice i, clipped normal noise with
        ``noise_param[i]`` will be applied respectively.

    Args:
        action: Raw action
        noise_param: Param of the normal noise.
        ratio: Sampled noise is multiplied with this ratio.

    Returns:
        Action with uniform noise.
    """
    if isinstance(noise_param[0], tuple):
        if len(noise_param) != action.shape[-1]:
            raise ValueError(
                "Noise param length doesn't match the last dimension of action")
        noise = t.rand(action.shape, device=action.device)
        for i in range(action.shape[-1]):
            noi_p = noise_param[i]
            noise.view(-1, noise.shape[-1])[:, i] *= noi_p[1] - noi_p[0]
            noise.view(-1, noise.shape[-1])[:, i] += noi_p[0]
            noise.view(-1, noise.shape[-1])[:, i].clamp(noi_p[2], noi_p[3])
    else:
        noise_param = noise_param  # type: Tuple[float, float, float, float]
        noise = t.rand(action.shape, device=action.device) \
            * (noise_param[1] - noise_param[0]) + noise_param[0]
        noise.clamp(noise_param[2], noise_param[3])
    return action + noise * ratio


def add_normal_noise_to_action(action: t.Tensor, noise_param=(0.0, 1.0),
                               ratio=1.0):
    """
    Add normal noise to action tensor.

    Hint:
        The innermost tuple contains:
        ``(normal_mean, normal_sigma)``

        If ``noise_param`` is ``Tuple[float, float]``,
        then the same normal noise will be added to ``action[*, :]``.

        If ``noise_param`` is ``Iterable[Tuple[float, float]]``,
        then for each ``action[*, i]`` slice i, clipped normal noise with
        ``noise_param[i]`` will be applied respectively.

    Args:
        action: Raw action
        noise_param: Param of the normal noise.
        ratio: Sampled noise is multiplied with this ratio.

    Returns:
        Action with normal noise.
    """
    if isinstance(noise_param[0], tuple):
        if len(noise_param) != action.shape[-1]:
            raise ValueError(
                "Noise range length doesn't match the last dimension of action")
        noise = t.randn(action.shape, device=action.device)
        for i in range(action.shape[-1]):
            noi_p = noise_param[i]
            noise.view(-1, noise.shape[-1])[:, i] *= noi_p[1]
            noise.view(-1, noise.shape[-1])[:, i] += noi_p[0]
    else:
        noise = t.rand(action.shape, device=action.device) \
                * noise_param[1] + noise_param[0]
    return action + noise * ratio


def add_ou_noise_to_action(action: t.Tensor, noise_param=(0.0, 1.0), ratio=1.0,
                           reset=False):
    """
    Add Ornstein-Uhlenbeck noise to action tensor.

    Warning:
        Ornstein-Uhlenbeck noise generator is shared. And only one set of noise
        param is supported.

    Args:
        action: Raw action
        noise_param: :class:`.OrnsteinUhlenbeckGen` params.
        ratio: Sampled noise is multiplied with this ratio.
        reset: Whether to reset the default Ornstein-Uhlenbeck noise generator.

    Returns:
        Action with Ornstein-Uhlenbeck noise.
    """
    global DEFAULT_OU_GEN
    if reset:
        DEFAULT_OU_GEN = None
    if DEFAULT_OU_GEN is None:
        DEFAULT_OU_GEN = OrnsteinUhlenbeckNoiseGen(action.shape,
                                                   *noise_param)
        DEFAULT_OU_GEN.reset()
    return action + DEFAULT_OU_GEN() * ratio



from machin.frame.buffers.prioritized_buffer import PrioritizedBuffer
# pylint: disable=wildcard-import, unused-wildcard-import
from .dqn import *


class RAINBOW(DQN):
    """
    RAINBOW DQN framework.
    """
    def __init__(self,
                 qnet: Union[NeuralNetworkModule, nn.Module],
                 qnet_target: Union[NeuralNetworkModule, nn.Module],
                 optimizer,
                 criterion,
                 value_min,
                 value_max,
                 *_,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
                 batch_size: int = 100,
                 update_rate: float = 0.005,
                 learning_rate: float = 0.001,
                 discount: float = 0.99,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 reward_func: Callable = None,
                 visualize: bool = False,
                 **__):
        """
        RAINBOW framework is described in
        `this <https://arxiv.org/abs/1710.02298>`__ essay.

        See Also:
            :class:`.DQN`

        Args:
            qnet: Q network module.
            qnet_target: Target Q network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            criterion: Criterion used to evaluate the value loss.
            value_min: Minimum of value domain.
            value_max: Maximum of value domain.
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`

            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            reward_func: Reward function used in training.
            mode: one of ``"vanilla", "fixed_target", "double"``.
            visualize: Whether visualize the network flow in the first pass.
        """
        super(RAINBOW, self).__init__(
            qnet, qnet_target, optimizer, criterion,
            learning_rate=learning_rate,
            lr_scheduler=lr_scheduler,
            lr_scheduler_args=lr_scheduler_args,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            batch_size=batch_size,
            update_rate=update_rate,
            discount=discount,
            replay_size=replay_size,
            replay_device=replay_device,
            replay_buffer=(PrioritizedBuffer(replay_size, replay_device)
                           if replay_buffer is None
                           else replay_buffer),
            reward_func=reward_func,
            visualize=visualize
        )
        self.v_min = value_min
        self.v_max = value_max

    def update(self,
               update_value=True,
               update_target=True,
               concatenate_samples=True,
               **__):
        # DOC INHERITED
        # pylint: disable=invalid-name
        (batch_size,
         (state, action, reward, next_state, terminal, *others),
         index, is_weight) = \
            self.replay_buffer.sample_batch(self.batch_size,
                                            concatenate_samples,
                                            sample_attrs=[
                                                "state", "action",
                                                "reward", "next_state",
                                                "terminal", "*"
                                            ])

        # q_dist is the distribution of q values,
        # with shape [batch_size, action_num, atom_num]
        q_dist = self.criticize(state)
        atom_num = q_dist.shape(-1)
        action["action"] = (action["action"]
                            .unsqueeze(-1)
                            .expand(batch_size, -1, atom_num))

        # support vector, shape: [atom_num]
        q_dist_support = t.linspace(self.v_min, self.v_max, atom_num)

        with t.no_grad():
            next_q_dist = self.criticize(next_state)
            target_next_q_dist = self.criticize(next_state, True)
            next_q_value = (next_q_dist * q_dist_support).sum(dim=2)
            next_action = t.max(next_q_value, dim=1)[1].unsqueeze(1)

            # shape: [batch_size, atom_num]
            target_next_q_dist = (target_next_q_dist
                                  .gather(dim=1, index=next_action)
                                  .squeeze(1))

        # shape: [batch_size, atom_num]
        q_dist_support = (q_dist_support
                          .unsqueeze(dim=0)
                          .expand_as(target_next_q_dist))

        # T_z is the bellman update for atom z_j
        # shape: [batch_size, atom_num]
        T_z = self.reward_func(reward.view, self.discount, q_dist_support,
                               terminal, *others)
        T_z = T_z.clamp(self.v_min, self.v_max)

        # delta_z is the interval length of each atom
        delta_z = (self.v_max - self.v_min) / (atom_num - 1.0)

        # b is the distance of T_z to v_min,
        # l and u are upper and lower atom indexes
        # b, l, u shape: [batch_size, atom_num]
        b = (T_z - self.v_min) / delta_z
        l, u = b.floor(), b.ceil()
        l_idx, l_dist = l.long().view(-1).cpu(), b - l
        u_idx, u_dist = u.long().view(-1).cpu(), u - b
        l_weight = (l_dist * target_next_q_dist).view(-1).cpu()
        u_weight = (u_dist * target_next_q_dist).view(-1).cpu()

        offset = (t.arange(0, batch_size * atom_num, atom_num)
                  .view([-1, 1])
                  .expand(batch_size, atom_num)
                  .view(-1))

        # distribute T_z probability to its nearest upper
        # and lower atom neighbors, using its distance to them.
        # shape: [batch_size, atom_num]
        # Note: index_add_ on CUDA is non-deterministic. will introduce noise
        target_dist = t.zeros_like(target_next_q_dist)
        target_dist.index_add_(dim=0, index=l_idx.view(-1) + offset,
                               source=l_weight)
        target_dist.index_add_(dim=0, index=u_idx.view(-1) + offset,
                               source=u_weight)

        # target_dist is equivalent to y_i in original dqn
        value_loss = -(target_dist * q_dist.log().cpu())
        value_loss = value_loss.sum(dim=1)

        abs_error = (t.abs(value_loss) + 1e-6).flatten().numpy()
        self.replay_buffer.update_priority(abs_error, index)

        value_loss = (value_loss *
                      t.from_numpy(is_weight).view([batch_size, 1])).mean()

        if self.visualize:
            self.visualize_model(value_loss, "qnet")

        if update_value:
            self.qnet.zero_grad()
            value_loss.backward()
            self.qnet_optim.step()

        # Update target Q network
        if update_target:
            soft_update(self.qnet_target, self.qnet, self.update_rate)

        # use .item() to prevent memory leakage
        return value_loss.item()



from typing import List
import inspect
import torch as t
import torch.nn as nn

from .helper_classes import Counter
from .tensor_board import SummaryWriter


def check_shape(tensor: t.Tensor, required_shape: List[int], name=""):
    """
    Check whether tensor has the specified shape.

    Args:
        tensor: Tensor to check.
        required_shape: A list of ``int`` specifying shape of each dimension.
        name: Name of tensor, will be printed in the error message.

    Raises:
        ``RuntimeError`` if shape of the tensor doesn't match.
    """
    shape = list(tensor.shape)
    if shape != required_shape:
        raise RuntimeError(
            "Tensor {} has invalid shape, required shape {}, is {}"
            .format(name, required_shape, shape)
        )


def check_nan(tensor: t.Tensor, name=""):
    """
    Check whether tensor has ``nan`` element.

    Args:
        tensor: Tensor to check
        name: Name of tensor, will be printed in the error message.

    Raises:
        ``RuntimeError`` if tensor has any ``nan`` element.
    """
    if t.any(t.isnan(tensor)):
        raise RuntimeError("Tensor {} contains nan!".format(name))


def _gen_input_check_hook(counter, interval, writer, hooks,
                          model, module_name):
    # Generate a input check hook which calls all sub hooks
    # when invoked by pytorch.
    def check_hook(module, input_):
        with t.no_grad():
            if counter.get() % interval == 0:
                # Get forward function signature.

                # Pytorch will not give us keyword arguments of modules,
                # and users also should not use keywork arguments in forward().
                # So we only need to get the 'args' part.
                input_names = inspect.getfullargspec(module.forward).args
                for input_name, input_value in zip(input_names, input_):
                    for hook in hooks:
                        hook(counter, writer, model, module,
                             module_name + ".input." + input_name,
                             input_value)
    return check_hook


def _gen_output_check_hook(counter, interval, writer, hooks,
                           model, module_name):
    # Generate a output check hook which calls all sub hooks
    # when invoked by pytorch.
    def check_hook(module, _input, output):
        with t.no_grad():
            if counter.get() % interval == 0:
                # Try to resolve output name, if failed, use
                # index as a substitute output name.
                # Currently, we can only judge output number if output is
                # a tuple.
                if isinstance(output, tuple):
                    default_names = [str(i) for i in range(len(output))]
                else:
                    default_names = ["0"]
                output_names = getattr(module, "_chk_output_names",
                                       default_names)
                for output_name, output_value in zip(output_names, output):
                    for hook in hooks:
                        hook(counter, writer, model, module,
                             module_name + ".output." + output_name,
                             output_value)
    return check_hook


def _gen_param_check_hook(counter, interval, writer, hooks,
                          model, module_name):
    # Generate a param check hook which calls all sub hooks
    # when invoked by pytorch.
    def check_hook(module, _grad_input, _grad_output):
        with t.no_grad():
            if counter.get() % interval == 0:
                for param_name, param_value in module.named_parameters():
                    for hook in hooks:
                        hook(counter, writer, model, module,
                             module_name + ".param." + param_name,
                             param_value)
    return check_hook


def i_chk_nan(_counter, _writer, _model, _module, input_name, input_val):
    """
    Check whether there is any nan element in the input, if input is a tensor.
    """
    if t.is_tensor(input_val):
        check_nan(input_val, input_name)


def i_chk_range(counter, writer, _model, _module, input_name, input_val):
    """
    Compute min, max and mean value of the input, if input is a tensor.
    """
    if t.is_tensor(input_val):
        writer.add_scalars(input_name,
                           {"min": t.min(input_val),
                            "max": t.max(input_val),
                            "mean": t.mean(input_val)},
                           counter.get())
        writer.flush()


def p_chk_nan(counter, _writer, _model, _module, param_name, param_val):
    """
    Check whether there is any nan element in the parameter.
    """
    check_nan(param_val,
              param_name + "(backward_count={})"
              .format(counter.get()))


def p_chk_range(counter, writer, _model, _module, param_name, param_val):
    """
    Compute min, max and mean value of the parameter.
    """
    writer.add_scalars(param_name,
                       {"min": t.min(param_val),
                        "max": t.max(param_val),
                        "mean": t.mean(param_val)},
                       counter.get())
    writer.add_histogram(param_name, param_val, counter.get())
    writer.flush()


def mark_as_atom_module(module):
    """
    Mark module as a atom leaf module, so it can be checked.
    """
    setattr(module, "_chk_is_atom", True)


def mark_module_output(module, output_names: List[str]):
    """
    Mark names of the module output. It will also tell checker
    about the number of outputs.

    Args:
        output_names: Name of each output value.
    """
    setattr(module, "_chk_output_names", output_names)


def check_model(writer: SummaryWriter,
                model: nn.Module,
                input_check_hooks=(i_chk_nan, i_chk_range),
                output_check_hooks=(),
                param_check_hooks=(p_chk_nan, p_chk_range),
                input_check_interval=1,
                output_check_interval=1,
                param_check_interval=100,
                name=""):
    """
    Check model input, output and parameters using hooks. Input and output
    check hooks are executed in the forward pass, and parameters are checked
    in the backward pass.

    An example::

        model = nn.Linear([100, 100])
        check_model(model)

        # Continue to do whatever you like.
        model(t.zeros([100]))

    Note:
        Only leaf modules will be checked (such as ``nn.Linear`` and not some
        complex neural network modules made of several sub-modules). But you
        can manually control granularity.

    Note:
        Param checkers might will be called multiple times if pytorch has to
        compute gradient on your model multiple times.

    Warning:
        Do not output ``tuple`` in your ``forward()`` function if you have
        output check hooks, otherwise you must specify names for each output.

    Hint:
        You may manually control the check granularity by using
        :func:`.mark_as_atom_module`.

        You may specify a list of names for your module outputs so
        names given to your output check hooks will not be numbers,
        by using :func:`.mark_module_output`

    Hint:
        For all three kinds of hooks, your hook need to have the following
        signature:

        ``hook(counter, writer, model, module, name, value)``

        where:

        - ``counter`` is the :class:`.Counter`, you can use
          :meth:`.Counter.get` to get the current pass number.
        - ``writer`` is :class:`.SummaryWriter` from ``tensorboardx``.
        - ``model`` is your model.
        - ``module`` is the module currently being checked.
        - ``name`` is input/output/parameter name string. For input, their
          detail names will be extracted from module ``forward`` signature.
          Output detail names will be numbers or names you have specified.
        - ``value`` is input/output/parameter value.

    Args:
        writer: Tensorboard ``SummaryWriter`` used to log.
        model: Model to be checked.
        input_check_hooks: A series of input check hooks.
        output_check_hooks: A series of output check hooks.
        param_check_hooks: A series of parameter check hooks.
        input_check_interval: Interval (number of forward passes)
            of input checking.
        output_check_interval: Interval (number of forward passes)
            of output checking.
        param_check_interval: Interval (number of backward passes)
            of parameter checking.
        name: Your model name.

    Returns:
        A function ``f()``, calling ``f()`` will deregister all check hooks.
    """

    forward_counter = Counter()
    backward_counter = Counter()

    def _forward_count(_, __):
        forward_counter.count()

    def _backward_count(_, __, ___):
        backward_counter.count()

    # These two hooks are used to move forward counters,
    # so that we can control check intervals.
    model.register_forward_pre_hook(_forward_count)
    model.register_backward_hook(_backward_count)

    handles = []

    # Register forward & backward checker hooks for all submodules.
    # Input checking are done in forward pre hooks.
    # Param checking are done in backward hooks.
    for sub_name, sub_module in model.named_modules(prefix=name):
        sub_module = sub_module  # type: nn.Module

        if (len(list(sub_module.modules())) != 1 and not
                getattr(sub_module, "_chk_is_atom", False)):
            # Current module has children, not a leaf module, so skip.
            continue

        handles.append(
            sub_module.register_forward_pre_hook(
                _gen_input_check_hook(forward_counter,
                                      input_check_interval,
                                      writer,
                                      input_check_hooks,
                                      model,
                                      sub_name)
            )
        )

        handles.append(
            sub_module.register_forward_hook(
                _gen_input_check_hook(forward_counter,
                                      output_check_interval,
                                      writer,
                                      output_check_hooks,
                                      model,
                                      sub_name)
            )
        )

        handles.append(
            sub_module.register_backward_hook(
                _gen_input_check_hook(forward_counter,
                                      param_check_interval,
                                      writer,
                                      param_check_hooks,
                                      model,
                                      sub_name)
            )
        )

    def cancel():
        for handle in handles:
            handle.remove()

    return cancel



"""
Attributes:
    default_logger: The default global logger.

TODO: maybe add logging utilities for distributed scenario?
"""
import colorlog
from logging import INFO

_default_handler = colorlog.StreamHandler()
_default_handler.setFormatter(colorlog.ColoredFormatter(
    "%(log_color)s[%(asctime)s] <%(levelname)s>:%(name)s:%(message)s"))

default_logger = colorlog.getLogger("default_logger")
default_logger.addHandler(_default_handler)
default_logger.setLevel(INFO)



import json
import argparse
from .helper_classes import Object


class Config(Object):
    """
    Attributes:
        restart_use_trial: Union[None, str]
    """
    def __init__(self):
        super(Config, self).__init__({
            "restart_use_trial": None
        })


def get_args():
    """
    Get arguments from the commandline.

    Note:
        ``--conf a=b`` will set ``<Returned Arg Object>.conf[a] = b``

        An example::

            python3 test.py --conf device=\"cuda:1\"
                            --conf some_dict={\"some_key\":1}

    Return:
         Argument object, each attribute is an argument, and
         ``.conf`` attribute is a ``dict``.
    """
    argparser = argparse.ArgumentParser(description=__doc__)
    argparser.add_argument(
        '-c', '--config_file',
        metavar='C',
        help='The Configuration file')
    argparser.add_argument("--conf", action="append")
    args = argparser.parse_args()

    config_dict = {}
    if args.config_file is not None:
        config_dict = load_config_dict(args.c)

    if args.conf is not None:
        for env_str in args.conf:
            name, value = env_str.split('=')
            value = eval(value)
            config_dict[name] = value
    args.conf = config_dict
    return args


def load_config_dict(json_file):
    """
    Get configs from a json file.

    Args:
        json_file: path to the json config file

    Return:
        config(dictionary)
    """
    # parse the configurations from the config json file provided
    with open(json_file, 'r') as config_file:
        config_dict = json.load(config_file)

    return config_dict


def merge_config(conf: Config, merge_conf: dict):
    """
    Merge config object with a dictionary, same keys in the config
    object will be overwritten.
    """
    for k, v in merge_conf.items():
        conf[k] = v


def replace_config(conf: Config, replace_conf: dict):
    """
    Clear the config object and replace its content with the dictionary.
    """
    conf.data.clear()
    for k, v in replace_conf.items():
        conf[k] = v


def save_config(conf: Config, json_file: str):
    """
    Dump config object to a json file.
    """
    with open(json_file, 'w') as config_file:
        json.dump(conf.data, config_file, sort_keys=True, indent=4)




from typing import Union
from abc import ABC, abstractmethod
import torch as t
import torch.nn as nn


class NeuralNetworkModule(nn.Module, ABC):
    """
    Note: input device and output device are determined by module parameters,
          your input module / output submodule should not store parameters on
          more than one device, and you also should not move your output to
          other devices other than your parameter storage device in forward().
    """
    def __init__(self):
        super(NeuralNetworkModule, self).__init__()
        self.input_module = None
        self.output_module = None

    def set_input_module(self, input_module: nn.Module):
        """
        Set the input submodule of current module.
        """
        self.input_module = input_module

        if not isinstance(input_module, NeuralNetworkModule):
            if isinstance(input_module, nn.Sequential):
                input_module = self.find_child(input_module, True)
            if len({p.device for p in input_module.parameters()}) > 1:
                raise RuntimeError("Input module must be another "
                                   "NeuralNetworkModule or locate "
                                   "on one single device.")

    def set_output_module(self, output_module: nn.Module):
        """
        Set the output submodule of current module.
        """
        self.output_module = output_module

        if not isinstance(output_module, NeuralNetworkModule):
            if isinstance(output_module, nn.Sequential):
                output_module = self.find_child(output_module, False)
            if len({p.device for p in output_module.parameters()}) > 1:
                raise RuntimeError("Output module must be another "
                                   "NeuralNetworkModule or locate "
                                   "on one single device.")

    @property
    def input_device(self):
        if self.input_module is None:
            raise RuntimeError("Input module not set.")
        else:
            if not isinstance(self.input_module, NeuralNetworkModule):
                dev_set = {p.device for p in self.input_module.parameters()}
                if len(dev_set) != 1:
                    raise RuntimeError("This input module contains "
                                       "parameters on different devices, "
                                       "please consider about splitting it.")
                else:
                    return list(dev_set)[0]
            else:
                return self.input_module.input_device

    @property
    def output_device(self):
        if self.output_module is None and self.input_module is None:
            raise RuntimeError("Output module not set.")
        elif self.output_module is not None:
            if not isinstance(self.output_module, NeuralNetworkModule):
                dev_set = {p.device for p in self.output_module.parameters()}
                if len(dev_set) != 1:
                    raise RuntimeError("This output module contains "
                                       "parameters on different devices, "
                                       "please consider about splitting it.")
                else:
                    return list(dev_set)[0]
            else:
                return self.output_module.output_device
        else:
            # input device is the same as output device
            return self.input_device

    @staticmethod
    def find_child(seq, is_first=True):
        """
        Find the first / last leaf child module.
        """
        if isinstance(seq, nn.Sequential):
            if is_first:
                return NeuralNetworkModule.find_child(seq[0], is_first)
            else:
                return NeuralNetworkModule.find_child(seq[-1], is_first)
        else:
            return seq

    @abstractmethod
    def forward(self, *_):
        pass


def static_module_wrapper(wrapped_module: nn.Module,
                          input_device: Union[str, t.device],
                          output_device: Union[str, t.device]):
    """
    Wrapped module could locate on multiple devices, but must not be moved.

    Input device and output device are statically specified by user.
    """
    wrapped_module.input_device = input_device
    wrapped_module.output_device = output_device
    return wrapped_module


def dynamic_module_wrapper(wrapped_module: nn.Module):
    """
    Wrapped module must locate on one single device, but could be moved around.

    Input device and output device are automatically detected.
    """
    wrapper = NeuralNetworkModule()
    wrapper.add_module("wrapped_module", wrapped_module)
    wrapper.set_input_module(wrapped_module)
    wrapper.set_output_module(wrapped_module)
    wrapper.forward = wrapped_module.forward
    return wrapper



from typing import Union, Dict, List, Tuple, Any, Callable
from ..transition import Transition, Scalar
import torch as t
import random


class Buffer(object):
    def __init__(self, buffer_size, buffer_device="cpu", *_, **__):
        """
        Create a buffer instance.

        Buffer stores a series of transition objects and functions
        as a ring buffer.

        See Also:
            :class:`.Transition`


        During sampling, the tensors in "state", "action" and "next_state"
        dictionaries, along with "reward", will be concatenated in dimension 0.
        any other custom keys specified in ``**kwargs`` will not be
        concatenated.

        Args:
            buffer_size: Maximum buffer size.
            buffer_device: Device where buffer is stored.
        """
        self.buffer_size = buffer_size
        self.buffer_device = buffer_device
        self.buffer = []
        self.index = 0

    def append(self, transition: Union[Transition, Dict],
               required_attrs=("state", "action", "next_state",
                               "reward", "terminal")):
        """
        Store a transition object to buffer.

        Args:
            transition: A transition object.
            required_attrs: Required attributes.

        Raises:
            ``ValueError`` if transition object doesn't have required
            attributes in ``required_attrs`` or has different attributes
            compared to other transition objects stored in buffer.
        """
        if isinstance(transition, dict):
            transition = Transition(**transition)
        if not transition.has_keys(required_attrs):
            missing_keys = set(required_attrs) - set(transition.keys())
            raise ValueError("Transition object missing attributes: {}"
                             .format(missing_keys))
        transition.to(self.buffer_device)

        if self.size() != 0 and self.buffer[0].keys() != transition.keys():
            raise ValueError("Transition object has different attributes!")

        if self.size() > self.buffer_size:
            # trim buffer to buffer_size
            self.buffer = self.buffer[(self.size() - self.buffer_size):]
        if self.size() == self.buffer_size:
            # ring buffer storage
            position = self.index
            self.buffer[self.index] = transition
            self.index += 1
            self.index %= self.buffer_size
        else:
            # append if not full
            self.buffer.append(transition)
            position = len(self.buffer) - 1
        return position

    def size(self):
        """
        Returns:
            Length of current buffer.
        """
        return len(self.buffer)

    def clear(self):
        self.buffer.clear()

    @staticmethod
    def sample_method_random_unique(buffer: List[Transition], batch_size: int) \
            -> Tuple[int, List[Transition]]:
        """
        Sample unique random samples from buffer.

        Note:
            Sampled size could be any value from 0 to ``batch_size``.
        """
        if len(buffer) < batch_size:
            batch = random.sample(buffer, len(buffer))
            real_num = len(buffer)
        else:
            batch = random.sample(buffer, batch_size)
            real_num = batch_size
        return real_num, batch

    @staticmethod
    def sample_method_random(buffer: List[Transition], batch_size: int) \
            -> Tuple[int, List[Transition]]:
        """
        Sample random samples from buffer.

        Note:
            Sampled size could be any value from 0 to ``batch_size``.
        """
        indexes = [random.randint(0, len(buffer) - 1)
                   for _ in range(batch_size)]
        batch = [buffer[i] for i in indexes]
        return batch_size, batch

    @staticmethod
    def sample_method_all(buffer: List[Transition], _) \
            -> Tuple[int, List[Transition]]:
        """
        Sample all samples from buffer. Always return the whole buffer.
        """
        return len(buffer), buffer

    def sample_batch(self,
                     batch_size: int,
                     concatenate: bool = True,
                     device: Union[str, t.device] = None,
                     sample_method: Union[Callable, str] = "random_unique",
                     sample_attrs: List[str] = None,
                     additional_concat_attrs: List[str] = None,
                     *_, **__) -> Any:
        """
        Sample a random batch from buffer.

        See Also:
            Default sample methods are defined as static class methods.

            :meth:`.Buffer.sample_method_random_unique`

            :meth:`.Buffer.sample_method_random`

            :meth:`.Buffer.sample_method_all`

        Note:
            "Concatenation"
            means ``torch.cat([...], dim=0)`` for tensors,
            and ``torch.tensor([...]).view(batch_size, 1)`` for scalars.

        Warnings:
            Custom attributes must not contain tensors. And only scalar custom
            attributes can be concatenated, such as ``int``, ``float``,
            ``bool``.

        Args:
            batch_size: A hint size of the result sample. actual sample size
                        depends on your sample method.
            sample_method: Sample method, could be one of:
                           ``"random", "random_unique", "all"``,
                           or a function:
                           ``func(list, batch_size)->(list, result_size)``
            concatenate: Whether concatenate state, action and next_state
                         in dimension 0.
                         If ``True``, for each value in dictionaries of major
                         attributes. and each value of sub attributes, returns
                         a concatenated tensor. Custom Attributes specified in
                         ``additional_concat_attrs`` will also be concatenated.
                         If ``False``, return a list of tensors.
            device:      Device to copy to.
            sample_attrs: If sample_keys is specified, then only specified keys
                         of the transition object will be sampled. You may use
                         ``"*"`` as a wildcard to collect remaining keys.
            additional_concat_attrs: additional custom keys needed to be
                         concatenated,

        Returns:
            Batch size, Sampled attribute values in the same order as
            ``sample_keys``.

            Sampled attribute values is a tuple. If batch size is zero,
            then ``None`` for each sampled keys.

            For major attributes, result are dictionaries of tensors with
            the same keys in your transition objects.

            For sub attributes, result are tensors.

            For custom attributes, if they are not in
            ``additional_concat_attrs``, then lists, otherwise tensors.
        """
        if isinstance(sample_method, str):
            if not hasattr(self, "sample_method_" + sample_method):
                raise RuntimeError("Cannot find specified sample method: {}"
                                   .format(sample_method))
            sample_method = getattr(self, "sample_method_" + sample_method)
        batch_size, batch = sample_method(self.buffer, batch_size)

        if device is None:
            device = self.buffer_device
        if sample_attrs is None:
            sample_attrs = batch[0].keys()
        if additional_concat_attrs is None:
            additional_concat_attrs = []

        return \
            batch_size, \
            self.post_process_batch(batch, device, concatenate,
                                    sample_attrs, additional_concat_attrs)

    @classmethod
    def post_process_batch(cls,
                           batch: List[Transition],
                           device: Union[str, t.device],
                           concatenate: bool,
                           sample_attrs: List[str],
                           additional_concat_attrs: List[str]):
        """
        Post-process (concatenate) sampled batch.
        """
        result = []
        used_keys = []
        if len(batch) == 0:
            return [None] * len(sample_attrs)

        major_attr = set(batch[0].major_attr)
        sub_attr = set(batch[0].sub_attr)
        for attr in sample_attrs:
            if attr in major_attr:
                tmp_dict = {}
                for sub_k in batch[0][attr].keys():
                    tmp_dict[sub_k] = cls.make_tensor_from_batch(
                        [item[attr][sub_k].to(device) for item in batch],
                        device, concatenate
                    )
                result.append(tmp_dict)
                used_keys.append(attr)
            elif attr in sub_attr:
                result.append(cls.make_tensor_from_batch(
                    [item[attr] for item in batch],
                    device, concatenate
                ))
                used_keys.append(attr)
            elif attr == "*":
                # select custom keys
                for remain_k in batch[0].keys():
                    if (remain_k not in major_attr and
                            remain_k not in sub_attr and
                            remain_k not in used_keys):
                        result.append(cls.make_tensor_from_batch(
                            [item[remain_k] for item in batch],
                            device,
                            concatenate and attr in additional_concat_attrs
                        ))
            else:
                result.append(cls.make_tensor_from_batch(
                    [item[attr] for item in batch],
                    device,
                    concatenate and attr in additional_concat_attrs
                ))
                used_keys.append(attr)
        return tuple(result)

    @staticmethod
    def make_tensor_from_batch(batch: List[Union[Scalar, t.Tensor]],
                               device: Union[str, t.device],
                               concatenate: bool):
        """
        Make a tensor from a batch of data.
        Will concatenate input tensors in dimension 0.
        Or create a tensor of size (batch_size, 1) for scalars.

        Args:
            batch: Batch data.
            device: Device to move data to
            concatenate: Whether performing concatenation.

        Returns:
            ``None`` if batch is empty,
            or tensor depends on your data (if concatenate),
            or original batch (if not concatenate).
        """
        if len(batch) == 0:
            return None
        if concatenate:
            item = batch[0]
            batch_size = len(batch)
            if t.is_tensor(item):
                batch = [it.to(device) for it in batch]
                return t.cat(batch, dim=0).to(device)
            else:
                return t.tensor(batch, device=device).view(batch_size, -1)
        else:
            return batch

    def __reduce__(self):
        # for pickling
        return self.__class__, (self.buffer_size, self.buffer_device)



from typing import Any, Union
from threading import Lock, Condition
from copy import deepcopy
from queue import Queue
from random import choice

import torch as t
import torch.nn as nn

from machin.parallel.distributed import RpcGroup
from machin.utils.prepare import prep_load_state_dict
from .ordered_server import OrderedServerBase, SimpleOrderedServer


class PushPullModelServer:
    """
    A simple parameter server, which synchronize model parameters
    by pushing and pulling all parameters and maintaining a strict
    ordered version chain.

    Warning:
        ``DistributedDataParallel`` is not supported.
    """
    MODEL_NAME = "_model"

    def __init__(self,
                 group: RpcGroup,
                 master: int = 0,
                 server: OrderedServerBase = None):
        """
        Args:
            group: Server group.
            master: The relative rpc rank of the master server in
                the default :class:`~machin.parallel.server.\
ordered_server.SimpleOrderedServer` implementation.
            server: Custom ordered server, mutually exclusive with
                ``master``.
        """
        self.server = (SimpleOrderedServer(group, master=master)
                       if server is None
                       else server)

    def push(self, model: nn.Module):
        """
        Try to push a model to the ordered server, if failed, the newest
        model will be automatically pulled and its parameters will be
        assigned to ``model``.

        Args:
            model: Model to push.
        """
        if not hasattr(model, "pp_version"):
            model.pp_version = 0
        model = deepcopy(model).to("cpu")

        if not self.server.push(
                self.MODEL_NAME, model.to("cpu").state_dict(),
                version=model.pp_version + 1, prev_version=model.pp_version
        ):
            result = self.server.pull(self.MODEL_NAME)
            if result is None:
                raise RuntimeError("Pull failed, this should not happen.")
            st_dict, version = result
            model.load_state_dict(st_dict)
            model.pp_version = version

    def pull(self, model: nn.Module):
        """
        Pull the newest state dict of your model and update its parameters
        and ``pp_version``.

        Args:
            model: Model to pull.
        """
        result = self.server.pull(self.MODEL_NAME)
        if result is None:
            return
        st_dict, version = result
        if not hasattr(model, "pp_version") or model.pp_version < version:
            prep_load_state_dict(model, st_dict)
            model.pp_version = version


class PushPullGradServer:
    """
    A simple parameter server, which synchronize model parameters
    by pushing gradients and pulling back new parameters, no strict
    order is guaranteed.

    Warning:
        ``DistributedDataParallel`` is not supported.
    """
    REDUCE_MASTER = 0
    REDUCE_SLAVE = 1

    def __init__(self,
                 group: RpcGroup,
                 reduce_master: int = 0,
                 model_master: int = 0,
                 server: OrderedServerBase = None,
                 reduce_device: Union[t.device, str] = "cpu",
                 reduce_batch_size: int = 64,
                 max_queue_size: int = 1024):
        """
        Note:
            Internally the master reducer will push updated versions
            to the ordered server group, this group does not provide
            ordering, and just provide model replication.

            Since a lot of workers will pull models concurrently (e.g.: A3C),
            we need to avoid the bottleneck of a single server providing
            the updated model to all these workers, therefore we need
            replication.

        Hint:
            Reduction is performed in a tree fashion:

            1. In the first step, clients will push new gradients to a
               random slave reducer, and the slave reducer will perform
               the first reduction pass, then slave reducers will push
               their results to the master reducer.
            2. In the second step, the master reducer will reduce results
               from the slave reducer to get the final reduced gradient
               dictionary (has the same structure as state_dict), and assign
               gradients to its **managed model**, and perform the
               optimization.
            3. In the final step, the master reducer will push the final
               model to the model server group, then clients can pull the
               newest model.

        Args:
            group: Server group.
            reduce_master: The relative rpc rank of the master reducer,
                which collects reduced gradients from slave reducers and
                perform the final reduction.
            model_master: The relative rpc rank of the master server in
                the default :class:`~machin.parallel.server.\
ordered_server.SimpleOrderedServer` implementation.
            server: Custom ordered server, mutually exclusive with
                ``master``.
            reduce_device: Device to perform reduction, by default it is "cpu".
            reduce_batch_size: Size of a single reduction batch, server will
                wait until the number of requests in the reduction queue have
                reached this size.
            max_queue_size: Maximum reduction request queue size.
        """
        self.server = (SimpleOrderedServer(group, master=model_master)
                       if server is None
                       else server)
        self.group = group
        self.group.rpc_register_paired(self.__class__, self)
        self.reduce_master = self.group.get_group_members()[reduce_master]
        self.reduce_slaves = self.group.get_group_members()
        self.reduce_batch_size = reduce_batch_size
        self.reduce_device = reduce_device
        self.master_queue = Queue(maxsize=max_queue_size)
        self.slave_queue = Queue(maxsize=max_queue_size)
        self.work_cond = Condition(Lock())
        self.model = None  # type: Union[nn.Module, None]
        self.optimizer = None

    def push(self, model: nn.Module):
        """
        Push the gradients of your model. Its gradients will be cleared.

        Args:
            model: Model to push.
        """
        # extract gradients from the model
        grad_dict = {}
        for k, v in model.parameters():
            grad_dict[k] = v.grad
        model.zero_grad()
        self.group.rpc_paired_class_async(
            choice(self.reduce_slaves),
            self._push_reply,
            self.__class__,
            args=(grad_dict, self.REDUCE_SLAVE)
        )

    def pull(self, model: nn.Module):
        """
        Pull the newest model.

        Args:
            model: Model to push.
        """
        params = self.server.pull("_push_pull_grad_managed_model")
        prep_load_state_dict(model, params)

    def manage_model(self, model: nn.Module, optimizer: Any):
        """
        Let the main reducer manage your model. Must be called before start.

        Args:
            model: Model to manage.
            optimizer: Optimizer of your model.

        Raises:
            ``RuntimeError`` if current rpc role is not the main reducer.
        """
        if self.group.get_cur_role() == self.reduce_master:
            self.model = model
            self.optimizer = optimizer
        else:
            raise RuntimeError("Current worker is not reduce master, and"
                               "cannot manage the model.")

    def _push_reply(self, grad_dict, level):
        # Append reduce requests to queue.
        if level == self.REDUCE_SLAVE:
            self.slave_queue.put_nowait(grad_dict)
            self.work_cond.notify_all()
        elif level == self.REDUCE_MASTER:
            self.master_queue.put_nowait(grad_dict)
            self.work_cond.notify_all()

    def _task_reduce_grad(self):
        while True:
            # Wait until one queue has reached target batch size
            self.work_cond.wait_for(
                lambda: (self.master_queue.qsize() > self.reduce_batch_size or
                         self.slave_queue.qsize() > self.reduce_batch_size)
            )
            if self.master_queue.qsize() > self.reduce_batch_size:
                # Perform reduction on the master reduction queue
                # Only the master reducer will execute this branch
                grad_dict = self._reduce_batch(self.master_queue,
                                               self.reduce_batch_size,
                                               self.reduce_device)
                # Assign gradients to the managed model and
                # perform optimization.
                if self.model is not None and self.optimizer is not None:
                    self.optimizer.zero_grad()
                    with t.no_grad():
                        for k, v in self.model.parameters():
                            v.grad = grad_dict[k].to(v.device)
                    self.optimizer.step()
                self.server.push("_push_pull_grad_managed_model",
                                 self.model.to("cpu").parameters(), 0, 0)

            elif self.slave_queue.qsize() > self.reduce_batch_size:
                # Perform reduction on the slave reduction queue
                # All processes(including master) in the reduction
                # group will execute this branch.
                grad_dict = self._reduce_batch(self.master_queue,
                                               self.reduce_batch_size,
                                               self.reduce_device)
                # Push reduced results to the master queue.
                self.group.rpc_paired_class_async(
                    self.reduce_master,
                    self._push_reply,
                    self.__class__,
                    args=(grad_dict, self.REDUCE_MASTER)
                )

    @staticmethod
    def _reduce_batch(queue, batch_size, reduce_device):
        """
        Perform batched gradient reduction

        Returns:
            Reduced gradient dictionary.
        """
        batch = []
        while len(batch) < batch_size:
            batch.append(queue.get())
        grad_dict = {}
        for grad in batch:
            for k, v in grad.items():
                if k not in grad_dict:
                    grad_dict[k] = [v.to(reduce_device)]
                else:
                    grad_dict[k].append(v.to(reduce_device))
        for k, v in grad_dict:
            # Stack parameter tensors in dim 0 and reduce.
            grad_dict[k] = t.mean(t.stack(v), dim=0, keepdim=True)
        return grad_dict



from typing import Dict
from torchviz import make_dot
import torch as t

from machin.utils.prepare import prep_load_model
from machin.utils.logging import default_logger


class TorchFramework:
    """
    Base framework for all algorithms
    """
    _is_top = []
    _is_restorable = []

    def __init__(self):
        self._visualized = set()

    @classmethod
    def get_restorable(cls):
        """
        Get restorable modules.
        """
        return cls._is_restorable

    def enable_multiprocessing(self):
        """
        Enable multiprocessing for all modules.
        """
        for top in self._is_top:
            model = getattr(self, top)
            model.share_memory()

    def load(self, model_dir: str, network_map: Dict[str, str] = None,
             version: int = -1):
        """
        Load models.

        An example of network map::

            {"restorable_model_1": "file_name_1",
             "restorable_model_2": "file_name_2"}

        Get keys by calling ``<Class name>.get_restorable()``

        Args:
            model_dir: Save directory.
            network_map: Key is module name, value is saved name.
            version: Version number of the save to be loaded.
        """
        network_map = {} if network_map is None else network_map
        restore_map = {}
        for r in self._is_restorable:
            if r in network_map:
                restore_map[network_map[r]] = getattr(self, r)
            else:
                default_logger.warning(
                    "Load path for module \"{}\" is not specified, "
                    "module name is used.".format(r),
                    RuntimeWarning
                )
                restore_map[r] = getattr(self, r)
        prep_load_model(model_dir, restore_map, version)

    def save(self, model_dir: str, network_map: Dict[str, str] = None,
             version: int = 0):
        """
        Save models.

        An example of network map::

            {"restorable_model_1": "file_name_1",
             "restorable_model_2": "file_name_2"}

        Get keys by calling ``<Class name>.get_restorable()``

        Args:
            model_dir: Save directory.
            network_map: Key is module name, value is saved name.
            version: Version number of the new save.
        """
        network_map = {} if network_map is None else network_map
        if version == -1:
            version = "default"
            default_logger.warning(
                "You are using the default version to save, "
                "use custom version instead.")
        for r in self._is_restorable:
            if r in network_map:
                t.save(getattr(self, r).state_dict(),
                       model_dir + "{}_{}.pt".format(network_map[r], version))
            else:
                default_logger.warning("Save name for module \"{}\" is not "
                                       "specified, module name is used."
                                       .format(r))
                t.save(getattr(self, r).state_dict(),
                       model_dir + "/{}_{}.pt".format(r, version))

    def visualize_model(self, final_tensor: t.Tensor, name: str):
        if name in self._visualized:
            return
        else:
            g = make_dot(final_tensor)
            g.view(filename=name + ".gv")
